#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to initiate a DataStore.
"""
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple, Union, Set, Any, Collection
import io
from warnings import warn
from collections import OrderedDict
import copy
import argparse
import shlex

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage
from ckanapi_harvesters.auxiliary.list_records import GeneralDataFrame
from ckanapi_harvesters.builder.builder_field import BuilderField
from ckanapi_harvesters.builder.mapper_datastore_multi import RequestFileMapperIndexKeys
from ckanapi_harvesters.harvesters.file_formats.file_format_abc import FileFormatABC
from ckanapi_harvesters.harvesters.file_formats.file_format_init import init_file_format_datastore
from ckanapi_harvesters.builder.mapper_datastore import DataSchemeConversion
from ckanapi_harvesters.builder.builder_resource import BuilderResourceABC, initial_resource_building_state
from ckanapi_harvesters.auxiliary.ckan_errors import DuplicateNameError
from ckanapi_harvesters.builder.builder_errors import RequiredDataFrameFieldsError, GroupByError
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo, CkanDataStoreInfo, CkanField, UpsertChoice
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, find_duplicates, datastore_id_col
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.auxiliary.ckan_configuration import datastore_default_upload_index_col_name, datastore_default_source_file_col_name, module_fields
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_init import init_data_cleaner


default_alias_keyword:Union[str,None] = "default"  # generate default alias if an alias with this value is found in parameters


class BuilderDataStoreABC(BuilderResourceABC, ABC):
    """
    The base class for DataStore resources. A DataStore resource can be updated with multiple requests and holds metadata for fields.

    :param field_builders: Merged metadata for fields (used in requests)
    :param field_builders_user: Field metadata specified by user (if exists, metadata from CKAN is prioritary)
    :param field_builders_data_source: Field metadata which could be obtained from the builder data source
    :param primary_key: primary key to transmit to CKAN (cannot be obtained through API)
    :param indexes: indexes to transmit to CKAN (cannot be obtained through API)
    :param aliases: Resource id aliases for requests (API cannot delete existing aliases)
    :param aux_upload_fun_name: Name of the function used to edit DataFrames before uploading
    :param aux_download_fun_name: Name of the function used to edit DataFrames after downloading
    :param aux_read_fun_name: Name of the function used to read file contents (defines local_file_format as a UserFileFormat)
    :param aux_write_fun_name: Name of the function used to write file contents (defines local_file_format as a UserFileFormat)
    :param local_file_format: Class used to read/write files
    :param df_mapper: DataFrame mapper function. This object adds certain indexes and applies the upload/download functions.
        It is responsible for mapping DataStore queries to file outputs.
    :param data_cleaner_upload: Data sanitizer used to automate certain tasks and replacing invalid values (default is None)
    """

    def __init__(self, *, parent, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, options_string:str=None, base_dir:str=None):
        super().__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id, download_url=download_url, options_string=options_string)
        self.field_builders: Union[Dict[str, BuilderField],None] = None
        self.field_builders_user: Union[Dict[str, BuilderField],None] = None
        self.field_builders_data_source: Union[Dict[str, BuilderField],None] = None
        # self.datastore_attributes: Union[CkanDataStoreInfo,None] = None
        # self.datastore_attributes_user: CkanDataStoreInfo = CkanDataStoreInfo()
        # self.datastore_attributes_data_source: Union[CkanDataStoreInfo,None] = None
        self.primary_key: Union[List[str],None] = None
        self.primary_key_user: Union[List[str],None] = None
        self.indexes: Union[List[str],None] = None
        self.aliases: Union[List[str],None] = None
        self.aux_upload_fun_name:str = ""
        self.aux_download_fun_name:str = ""
        self.aux_read_fun_name:str = ""
        self.aux_write_fun_name:str = ""
        # Functions input/outputs
        self.data_cleaner_upload: Union[CkanDataCleanerABC,None] = None
        self.reupload_on_update = False  # do not reupload on update for DataStores
        self.reupload_if_needed: bool = True
        self.reupload_needed: Union[bool,None] = None
        self.df_mapper = DataSchemeConversion()
        self.local_file_format: Union[FileFormatABC,None] = None
        self.column_enable_upload_index: bool = True
        self.column_enable_source_file: bool = False
        self.read_line_counter:int = 0
        self.upload_start_line:int = 0

    def copy(self, *, dest=None, parent=None):
        super().copy(dest=dest, parent=parent)
        dest.field_builders = copy.deepcopy(self.field_builders)
        dest.field_builders_user = copy.deepcopy(self.field_builders_user)
        dest.field_builders_data_source = copy.deepcopy(self.field_builders_data_source)
        dest.primary_key = copy.deepcopy(self.primary_key)
        dest.primary_key_user = copy.deepcopy(self.primary_key_user)
        dest.indexes = copy.deepcopy(self.indexes)
        dest.aliases = copy.deepcopy(self.aliases)
        dest.aux_upload_fun_name = self.aux_upload_fun_name
        dest.aux_download_fun_name = self.aux_download_fun_name
        dest.aux_read_fun_name = self.aux_read_fun_name
        dest.aux_write_fun_name = self.aux_write_fun_name
        dest.reupload_on_update = self.reupload_on_update
        dest.reupload_if_needed = self.reupload_if_needed
        dest.reupload_needed = self.reupload_needed
        dest.column_enable_upload_index = self.column_enable_upload_index
        dest.column_enable_source_file = self.column_enable_source_file
        dest.df_mapper = self.df_mapper.copy()
        dest.local_file_format = self.local_file_format.copy()
        return dest

    def _update_metadata(self, ckan: CkanApi, *, base_dir:str=None) -> None:
        super()._update_metadata(ckan=ckan, base_dir=base_dir)

    @staticmethod
    def _setup_cli_parser(parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="DataStore resource specific options", add_help=False,
                                             epilog=
                                             "Examples: \n"
                                             "- Selecting a Data Cleaner: --data-cleaner GeoJSON \n"
                                             "- Process one file per primary key combination (first columns of the primary key, except the last one): --one-frame-per-primary-key --no-chunks")
        parser.add_argument("--data-cleaner", type=str,
                            help="Data cleaner to call before uploading data")
        parser.add_argument("--one-frame-per-primary-key",
                            help="Enabling this option makes the upload process expect one DataFrame per primary key combination (except the last field of the primary key, which could be an index in the file).\n"
                            "This option should be associated with the file format option --no-chunks to ensure a file is treated at once", action="store_true", default=False)
        parser.add_argument("--group-by", type=str,
                            help="Fields of the primary key defining the request to reconstruct a file, in --one-frame-per-primary-key mode, separated by a comma (no spaces). \n"
                                 "By default, the first columns of the primary, except the last one is used. \n"
                                 "At least one field of the primary key must be unused here. \n")
        parser.add_argument("--include-source-file",
                            help="Add a column for the file name of the original data (named " + datastore_default_source_file_col_name + ")", action="store_true", default=False)
        parser.add_argument("--no-upload-index",
                            help="Disable the generation of an upload index column in case no primary key was given (named " + datastore_default_upload_index_col_name + ")", action="store_true", default=False)
        return parser

    def _setup_cli_parser_external(self, parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        # return FileFormatABC._setup_cli_parser(parser)
        return parser

    def print_help_cli(self, display:bool=True) -> str:
        parser = self._setup_cli_parser()
        self._setup_cli_parser_external(parser)
        if display:
            parser.print_help()
        with io.StringIO() as stream:
            parser.print_help(stream)
            return stream.getvalue()

    def _cli_args_apply(self, args: argparse.Namespace, *, base_dir: str = None, error_not_found: bool = True) -> None:
        if args.data_cleaner is not None:
            self.data_cleaner_upload = init_data_cleaner(args.data_cleaner)
        if args.no_upload_index:
            self.column_enable_upload_index = False
        if args.include_source_file:
            self.column_enable_source_file = True
        self.setup_default_file_mapper()
        if args.one_frame_per_primary_key is not None and args.one_frame_per_primary_key:
            self.apply_one_frame_per_primary_key(args.group_by)
        elif args.group_by is not None:
            msg = GroupByError("Argument --group-by cannot be used without option --one-frame-per-primary-key")
            warn(msg)

    def setup_default_file_mapper(self, *, primary_key:List[str]=None, file_query_list:Collection[Tuple[str, dict]]=None) -> None:
        if primary_key is None:
            if self.primary_key_user is not None:
                self.primary_key = self.primary_key_user
        else:
            self.primary_key = primary_key
        if self.column_enable_source_file:
            self.df_mapper.source_file_column = datastore_default_source_file_col_name
        else:
            self.df_mapper.source_file_column = ""

    def apply_one_frame_per_primary_key(self, group_by_argument:Union[str, List[str]]=None):
        """
        Enables mode --one-frame-per-primary-key
        and applies option --group-by

        In this mode, the upload process expect one DataFrame per primary key combination
        (except the last field of the primary key, which could be an index in the file).
        Upload update checks are performed using this assumption (do not read files by chunks).
        Downloads fill files according to unique combinations of the first columns of the primary key.
        """
        assert (self.primary_key is not None and len(self.primary_key) > 1)
        if group_by_argument is None:
            # default mode
            group_by_keys = self.primary_key[:-1]
            # sort_by_keys = self.primary_key[-1:]
            sort_by_keys = self.primary_key
        else:
            # custom --group-by argument
            if isinstance(group_by_argument, str):
                group_by_keys = group_by_argument.split(ckan_tags_sep)
            else:
                group_by_keys = group_by_argument
            group_by_keys =  [key.strip() for key in group_by_keys]
            extra_keys_primary = set(group_by_keys) - set(self.primary_key)
            if len(extra_keys_primary) > 0:
                raise GroupByError("--group-by argument must contain only columns of the primary key")
            elif len(group_by_keys) == len(self.primary_key):
                raise GroupByError("--group-by argument must not contain all columns of the primary key")
            # sort_by_keys = list(set(self.primary_key) - set(group_by_keys))
            sort_by_keys = self.primary_key
        self.df_mapper = RequestFileMapperIndexKeys(group_by_keys=group_by_keys, sort_by_keys=sort_by_keys)
        if self.local_file_format.read_by_chunks_enabled():
            msg = "Mode --one-frame-per-primary-key is not compatible with reading files by chunks. Disabling this feature."
            warn(msg)
            self.local_file_format.allow_chunks = False

    def initialize_extra_options_string(self, extra_options_string:str, base_dir:str) -> None:
        self.local_file_format = init_file_format_datastore(self.resource_attributes_user.format, extra_options_string, self.aux_read_fun_name, self.aux_write_fun_name)  # default file format is CSV (user can change)

    def _merge_resource_attributes_from_file(self) -> None:
        """
        This function merges metadata which could have been extracted from a file reading function into the attributes from data source.
        Call after self.local_file_format.read_file()
        """
        resource_attributes_from_file = self.local_file_format.resource_attributes_from_file
        if resource_attributes_from_file is not None:
            if self.resource_attributes_data_source is None:
                self.resource_attributes_data_source = resource_attributes_from_file
            else:
                self.resource_attributes_data_source.update_missing(resource_attributes_from_file)
                if resource_attributes_from_file.datastore_info is not None and resource_attributes_from_file.datastore_info.fields_dict is not None:
                    if self.resource_attributes_data_source.datastore_info is None:
                        self.resource_attributes_data_source.datastore_info = resource_attributes_from_file.datastore_info
                    elif self.resource_attributes_data_source.datastore_info.fields_dict is None:
                        self.resource_attributes_data_source.datastore_info.fields_dict = resource_attributes_from_file.datastore_info.fields_dict
                    for field_name, field_info in resource_attributes_from_file.datastore_info.fields_dict.items():
                        field_builder = BuilderField._from_ckan_field(field_info)
                        if field_name in self.field_builders_data_source.keys():
                            self.field_builders_data_source[field_name].update_missing(field_builder)
                        else:
                            self.field_builders_data_source[field_name] = field_builder

    def initialize_from_options_string(self, base_dir:str, *, options_string:str=None, parser:argparse.ArgumentParser=None) -> None:
        if options_string is None:
            options_string = self.options_string
        if options_string is None:
            self.initialize_extra_options_string(None, base_dir=base_dir)
            return
        parser = self._setup_cli_parser(parser)
        args, extra_args = parser.parse_known_args(shlex.split(options_string))
        self.initialize_extra_options_string(shlex.join(extra_args), base_dir=base_dir)
        self._cli_args_apply(args)

    def init_options_from_ckan(self, ckan:CkanApi, *, base_dir:str=None) -> None:
        if self.known_resource_info is None:
            self.known_resource_info = ckan.get_resource_info_or_request(self.name, self.package_name,
                                                                         error_not_found=False, datastore_info=True)
        super().init_options_from_ckan(ckan, base_dir=base_dir)
        # self.local_file_format.chunk_size = ckan.params.default_limit_read
        if self.field_builders is not None:
            for field_builder in self.field_builders.values():
                field_builder.internal_attrs.update_from_ckan(ckan)

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        primary_keys_string: Union[str,None] = _string_from_element(row["primary key"])
        indexes_string: Union[str,None] = _string_from_element(row["indexes"])
        aliases_string: Union[str,None] = None
        if "data cleaner" in row.keys():
            data_cleaner_string = _string_from_element(row["data cleaner"], empty_value="")
            if data_cleaner_string is not None:
                self.data_cleaner_upload = init_data_cleaner(data_cleaner_string)
        if "upload function" in row.keys():
            self.aux_upload_fun_name: str = _string_from_element(row["upload function"], empty_value="")
        if "download function" in row.keys():
            self.aux_download_fun_name: str = _string_from_element(row["download function"], empty_value="")
        if "read function" in row.keys():
            self.aux_read_fun_name: str = _string_from_element(row["read function"], empty_value="")
        if "write function" in row.keys():
            self.aux_write_fun_name: str = _string_from_element(row["write function"], empty_value="")
        if "aliases" in row.keys():
            aliases_string = _string_from_element(row["aliases"])
        if primary_keys_string is not None:
            if primary_keys_string.lower() == "none":
                self.primary_key_user = []
            else:
                self.primary_key_user = [field.strip() for field in primary_keys_string.split(ckan_tags_sep)]
        if indexes_string is not None:
            if indexes_string.lower() == "none":
                self.indexes = []
            else:
                self.indexes = [field.strip() for field in indexes_string.split(ckan_tags_sep)]
        if aliases_string is not None:
            self.aliases = aliases_string.split(ckan_tags_sep)
        self.setup_default_file_mapper()
        self.initialize_from_options_string(base_dir=base_dir)

    @abstractmethod
    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["Primary key"] = ckan_tags_sep.join(self.primary_key_user) if self.primary_key_user else ""
        d["Indexes"] = ckan_tags_sep.join(self.indexes) if self.indexes is not None else ""
        d["Data cleaner"] = self.data_cleaner_upload.get_class_keyword() if self.data_cleaner_upload is not None else ""
        d["Upload function"] = self.aux_upload_fun_name
        d["Download function"] = self.aux_download_fun_name
        d["Read function"] = self.aux_read_fun_name
        d["Write function"] = self.aux_write_fun_name
        d["Aliases"] = ckan_tags_sep.join(self.aliases) if self.aliases is not None else ""
        return d

    def _check_field_duplicates(self):
        if self.field_builders is not None:
            duplicates = find_duplicates([field_builder.name for field_builder in self.field_builders.values()])
            if len(duplicates) > 0:
                raise DuplicateNameError("Field", duplicates)

    def _get_fields_dict(self) -> Dict[str, dict]:
        self._check_field_duplicates()
        # self._get_fields_update(ckan=None, current_df_fields=None)
        if self.field_builders_user is not None:
            fields_dict = OrderedDict([(field_builder.name, field_builder._to_dict()) for field_builder in self.field_builders_user.values()])
        else:
            fields_dict = None
        return fields_dict

    def _get_fields_info(self) -> Dict[str, CkanField]:
        self._check_field_duplicates()
        if self.field_builders_user is not None:
            builder_fields = OrderedDict([(field_builder.name, field_builder._to_ckan_field()) for field_builder in self.field_builders_user.values()])
        else:
            builder_fields = {}
        return builder_fields

    def _get_fields_df(self) -> Union[pd.DataFrame,None]:
        if self.field_builders_user is not None:
            fields_dict_list = [value for value in self._get_fields_dict().values()]
            fields_df = pd.DataFrame.from_records(fields_dict_list)
            return fields_df
        else:
            return None

    def _load_fields_df(self, fields_df: pd.DataFrame):
        fields_df.columns = fields_df.columns.map(str.lower)
        fields_df.columns = fields_df.columns.map(str.strip)
        self.field_builders_user = OrderedDict()
        for row_loc, row in fields_df.iterrows():
            field_builder = BuilderField()
            field_builder._load_from_df_row(row=row)
            if field_builder.name in self.field_builders_user.keys():
                raise DuplicateNameError("field_builder", [field_builder.name])
            self.field_builders_user[field_builder.name] = field_builder

    def _to_ckan_resource_info(self, package_id:str, check_id:bool=True) -> CkanResourceInfo:
        resource_info = super()._to_ckan_resource_info(package_id=package_id, check_id=check_id)
        resource_info.datastore_info = CkanDataStoreInfo()
        resource_info.datastore_info.resource_id = resource_info.id
        if self.field_builders is not None:
            resource_info.datastore_info.fields_dict = OrderedDict()
            for name, field_builder in self.field_builders.items():
                resource_info.datastore_info.fields_dict[name] = field_builder._to_ckan_field()
        else:
            resource_info.datastore_info.fields_dict = None
        resource_info.datastore_info.fields_id_list = [name for name, field_builder in self.field_builders.items()] if self.field_builders is not None else []
        if self.indexes is not None:
            resource_info.datastore_info.index_fields = self.indexes.copy()
        aliases = self._get_alias_list(None)
        if aliases is not None:
            resource_info.datastore_info.aliases = aliases.copy()
        return resource_info

    @abstractmethod
    def load_sample_df(self, resources_base_dir:str, *, upload_alter:bool=True) -> GeneralDataFrame:
        """
        Function returning the data from the indicated resources as a pandas DataFrame.
        This is the DataFrame equivalent for load_sample_data.

        :param resources_base_dir: base directory to find the resources on the local machine
        :return:
        """
        raise NotImplementedError()

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return False

    def get_sample_file_path(self, resources_base_dir: str, ckan:Union[CkanApi,None]=None) -> None:
        return None

    def load_sample_data(self, resources_base_dir:str) -> bytes:
        df = self.load_sample_df(resources_base_dir=resources_base_dir)
        return self.local_file_format.write_in_memory(df, fields=self._get_fields_info())

    def upsert_request_df(self, ckan: CkanApi, df_upload:pd.DataFrame, *,
                          total_lines_read:int, file_name:str,
                          method:UpsertChoice=UpsertChoice.Upsert,
                          apply_last_condition:bool=None, always_last_condition:bool=None) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Call to ckan datastore_upset.
        Before sending the DataFrame, a call to df_upload_alter is made.
        This method is overloaded in BuilderDataStoreMultiABC and BuilderDataStoreFolder

        :param ckan:
        :param df_upload:
        :param method:
        :return:
        """
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=True)
        df_upload_transformed = self.df_mapper.df_upload_alter(df_upload, total_lines_read=total_lines_read,
                                                               fields=self._get_fields_info(), file_query=file_name)
        ret_df = ckan.datastore_upsert(df_upload_transformed, resource_id, method=method,
                                       apply_last_condition=apply_last_condition,
                                       always_last_condition=always_last_condition,
                                       data_cleaner=self.data_cleaner_upload,
                                       progress_callback=self.progress_callback)
        return df_upload_transformed, ret_df

    def upsert_request_final(self, ckan: CkanApi, *, force:bool=False) -> None:
        """
        Final steps after the last upsert query.
        These steps are automatically done for a DataStore defined by one file.

        :param ckan:
        :param force: perform request anyways
        :return:
        """
        if force:
            resource_id = self.get_or_query_resource_id(ckan, error_not_found=True)
            ckan.datastore_upsert_last_line(resource_id=resource_id)

    def _get_alias_list(self, ckan:Union[CkanApi,None]):
        aliases = self.aliases
        if default_alias_keyword is not None:
            if ckan is not None:
                default_alias_name = ckan.datastore_default_alias(self.name, self.package_name, error_not_found=False)
            else:
                default_alias_name = CkanApi.datastore_default_alias_of_names(self.name, self.package_name)
            if aliases is not None:
                for i, alias in enumerate(aliases):
                    if alias.lower().strip() == default_alias_keyword:
                        aliases[i] = default_alias_name
        return aliases

    def _check_necessary_fields(self, current_fields: Set[str] = None, empty_datastore:bool=False, raise_error: bool = True) -> Set[str]:
        """
        Auxiliary function to list the fields which are required:
        - for df_mapper to determine the file names, associated requests, and recognize the last inserted row of a document.
        - to initialize the DataStore with the columns for the primary key and indexes

        The required fields are compared to current_fields, if provided.
        """
        if empty_datastore:
            return set()
        required_fields = self.df_mapper.get_necessary_fields()
        if self.primary_key is not None:
            required_fields = required_fields.union(set(self.primary_key))
        if self.indexes is not None:
            required_fields = required_fields.union(set(self.indexes))
        if current_fields is not None:
            missing_fields = required_fields - current_fields
            if len(missing_fields) > 0:
                msg = RequiredDataFrameFieldsError(missing_fields)
                if raise_error:
                    raise msg
                else:
                    warn(str(msg))
        return required_fields

    def _check_undocumented_fields(self, current_fields: Set[str]) -> None:
        if self.field_builders_user is not None:
            # list fields which are not documented
            fields_doc = set(self.field_builders_user.keys())
            missing_doc = current_fields - fields_doc - module_fields
            extra_doc = fields_doc - current_fields
            if len(extra_doc) > 0:
                msg = f"{len(extra_doc)} extra fields were documented but absent of sample data for table {self.name}: {', '.join(extra_doc)}"
                warn(msg)
            if len(missing_doc) > 0:
                msg = f"{len(missing_doc)} fields are left documented for table {self.name}: {', '.join(missing_doc)}"
                warn(msg)
        else:
            msg = f"No field documentation was provided for table {self.name}. {len(current_fields)} fields are left documented: {', '.join(current_fields)}"
            warn(msg)

    def _get_fields_update(self, ckan: CkanApi, *, current_df_fields:Union[Set[str],None], data_cleaner_fields:Union[List[dict],None],
                           reupload:bool, override_ckan:bool) -> OrderedDict[str, CkanField]:
        """
        Merge field builders in the following order of priority:
        1. Existing metadata from CKAN (can be ignored with option override_ckan)
        2. Metadata specified by the user in the Excel worksheet
        3. Metadata found automatically from the data source (e.g. in file header or database)
        4. Metadata found automatically by the data cleaner, especially for field typing
        """
        # if current_df_fields is not None:
        #     current_df_fields_lower = {field_name.lower() for field_name in current_df_fields}
        # else:
        #     current_df_fields_lower = None
        self.field_builders = OrderedDict()
        allowed_fields = current_df_fields
        # 1. Information specified by user in Excel, for preserving field specified order
        impose_user_fields = False
        if self.field_builders_user is not None:
            for field_name, field_builder in self.field_builders_user.items():
                self.field_builders[field_name] = field_builder.copy()
            if impose_user_fields:
                if allowed_fields is not None:
                    allowed_fields = allowed_fields.union(set(self.field_builders_user.keys()))
                else:
                    allowed_fields = set(self.field_builders_user.keys())
        # 2. Current CKAN information: update missing information
        if (not override_ckan) and self.known_resource_info is not None and self.known_resource_info.datastore_info is not None and self.known_resource_info.datastore_info.fields_dict is not None:
            known_fields = self.known_resource_info.datastore_info.fields_dict
            for field_name, field_info in known_fields.items():
                if current_df_fields is not None and field_name in current_df_fields:
                    if field_name in self.field_builders.keys():
                        self.field_builders[field_name].update_missing(BuilderField._from_ckan_field(field_info))
                    else:
                        self.field_builders[field_name] = BuilderField._from_ckan_field(field_info)
        # 3. Metadata generated from data source: update missing information
        if self.field_builders_data_source is not None:
            for field_name, field_builder in self.field_builders_data_source.items():
                if field_name in self.field_builders.keys():
                    self.field_builders[field_name].update_missing(field_builder)
                else:
                    self.field_builders[field_name] = field_builder.copy()
        # 4. Metadata generated from Data cleaner
        if data_cleaner_fields is not None:
            for field_dict in data_cleaner_fields:
                field_builder = BuilderField._from_ckan_field(CkanField.from_ckan_dict(field_dict))
                if field_builder.name in self.field_builders.keys():
                    self.field_builders[field_builder.name].update_missing(field_builder)
                else:
                    self.field_builders[field_builder.name] = field_builder.copy()
        if self.field_builders is not None:
            if current_df_fields is not None:
                builder_fields = [field_builder._to_ckan_field() for field_builder in self.field_builders.values() if field_builder.name in allowed_fields]
            else:
                # use case: get all known fields (before data_cleaner)
                builder_fields = [field_builder._to_ckan_field() for field_builder in self.field_builders.values()]
        else:
            builder_fields = None
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=False)
        if resource_id is not None and not reupload:
                # This call merges the field builder with existing information
            update_needed, fields_update = ckan.datastore_field_patch_dict(fields_merge=data_cleaner_fields, fields_update=builder_fields,
                                                                           return_list=False, datastore_merge=not override_ckan,
                                                                           resource_id=resource_id, error_not_found=False)
        else:
            fields_update = CkanApi.datastore_field_dict(fields_merge=data_cleaner_fields, fields_update=builder_fields, return_list=False)
        return fields_update

    def _collect_indexes_from_fields(self) -> Set[str]:
        if self.field_builders is not None:
            return {field_builder.name for field_builder in self.field_builders.values() if field_builder.is_index}
        else:
            return set()

    def _get_primary_key_indexes(self, data_cleaner_index: Set[str], current_df_fields:Set[str], error_missing:bool, empty_datastore:bool=False) -> Tuple[Union[List[str],None], Union[List[str],None]]:
        # update primary keys and indexes: only if present
        if empty_datastore:
            return None, None
        primary_key = None
        if current_df_fields is None:
            primary_key = self.primary_key
        elif self.primary_key is not None:
            extra_primary_key = set(self.primary_key) - current_df_fields
            if len(extra_primary_key) == 0:
                primary_key = self.primary_key
            elif error_missing:
                raise RequiredDataFrameFieldsError(extra_primary_key)
        indexes = None
        if self.indexes is not None:
            indexes_full_set = set(self.indexes).union(self._collect_indexes_from_fields()).union(data_cleaner_index)
        else:
            indexes_full_set = self._collect_indexes_from_fields().union(data_cleaner_index)
        if primary_key is not None:
            indexes_full_set = indexes_full_set - set(primary_key)
        if len(indexes_full_set) == 0:
            indexes_full = None
        else:
            indexes_full = list(indexes_full_set)
        if current_df_fields is None:
            indexes = indexes_full
        elif indexes_full is not None:
            extra_indexes = set(indexes_full) - current_df_fields
            if len(extra_indexes) == 0:
                indexes = indexes_full
            elif error_missing:
                raise RequiredDataFrameFieldsError(extra_indexes)
        return primary_key, indexes

    def _compare_fields_to_datastore_info(self, resource_info:CkanResourceInfo, current_fields: Set[str], ckan:CkanApi) -> None:
        # compare fields with DataStore info (if present, for information)
        if resource_info.datastore_info is not None:
            fields_info = set(resource_info.datastore_info.fields_id_list)
            missing_info = current_fields - fields_info
            extra_info = fields_info - current_fields
            if len(extra_info) > 0:
                msg = f"{len(extra_info)} extra fields are in the database but absent of sample data for table {self.name}: {', '.join(extra_info)}"
                warn(msg)
            if len(missing_info) > 0 and ckan.params.verbose_request:
                msg = f"{len(missing_info)} fields are not in DataStore info because they are being added for table {self.name}: {', '.join(missing_info)}"
                print(msg)

    def _apply_data_cleaner_before_patch(self, ckan:CkanApi, df_upload: pd.DataFrame, *, reupload:bool, override_ckan:bool) -> Tuple[pd.DataFrame, List[dict], Set[str]]:
        if df_upload is not None and self.data_cleaner_upload is not None:
            fields_for_cleaner = self._get_fields_update(ckan, current_df_fields=None, data_cleaner_fields=None, reupload=reupload, override_ckan=override_ckan)
            df_upload = self.data_cleaner_upload.clean_records(df_upload, known_fields=fields_for_cleaner, inplace=True)
            data_cleaner_fields = self.data_cleaner_upload.merge_field_changes(fields_for_cleaner)
            data_cleaner_index = self.data_cleaner_upload.field_suggested_index
        else:
            data_cleaner_fields = None
            data_cleaner_index = set()
        return df_upload, data_cleaner_fields, data_cleaner_index

    def patch_request(self, ckan: CkanApi, package_id: str, *,
                      df_upload: pd.DataFrame=None, reupload: bool = None, override_ckan:bool=False,
                      resources_base_dir:str=None) -> CkanResourceInfo:
        self._merge_resource_attributes(override_ckan=override_ckan)
        if reupload is None: reupload = self.reupload_on_update
        if df_upload is None:
            df_upload = self.load_sample_df(resources_base_dir=resources_base_dir, upload_alter=True)
        else:
            pass  # do not alter df_upload because it should already be in the database format
        df_upload, data_cleaner_fields, data_cleaner_index = self._apply_data_cleaner_before_patch(ckan, df_upload, reupload=reupload, override_ckan=override_ckan)
        current_df_fields = set(df_upload.columns) - {datastore_id_col}  # _id field cannot be documented
        empty_datastore = df_upload is None or len(df_upload) == 0
        self._check_necessary_fields(current_df_fields, empty_datastore=empty_datastore, raise_error=True)
        self._check_undocumented_fields(current_df_fields)
        aliases = self._get_alias_list(ckan)
        primary_key, indexes = self._get_primary_key_indexes(data_cleaner_index, current_df_fields=current_df_fields,
                                                             error_missing=True, empty_datastore=empty_datastore)
        fields_update = self._get_fields_update(ckan, current_df_fields=current_df_fields,
                                                data_cleaner_fields=data_cleaner_fields, reupload=reupload, override_ckan=override_ckan)
        fields = list(fields_update.values()) if len(fields_update) > 0 else None
        resource_info = ckan.resource_create(package_id, name=self.name, format=self.resource_attributes.format,
                                             description=self.resource_attributes.description,
                                             state=initial_resource_building_state if initial_resource_building_state is not None else self.resource_attributes.state,
                                             create_default_view=self.create_default_view,
                                             cancel_if_exists=True, update_if_exists=True, reupload=reupload,
                                             datastore_create=True, records=df_upload, fields=fields,
                                             primary_key=primary_key, indexes=indexes, aliases=aliases,
                                             progress_callback=self.progress_callback)
        resource_id = resource_info.id
        self.known_id = resource_id
        self._compare_fields_to_datastore_info(resource_info, current_df_fields, ckan)
        return resource_info

    def download_resource_df(self, ckan: CkanApi, search_all:bool=True, download_alter:bool=True, **kwargs) -> Union[pd.DataFrame,None]:
        """
        Download the resource and return it as a DataFrame.
        This is the DataFrame equivalent for download_resource_bytes.

        :param ckan:
        :param search_all:
        :param download_alter:
        :param kwargs:
        :return:
        """
        resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=self.download_error_not_found)
        if resource_id is None and not self.download_error_not_found:
            return None
        df_download = ckan.datastore_dump(resource_id, search_all=search_all, progress_callback=self.progress_callback, **kwargs)
        if download_alter:
            df_local = self.df_mapper.df_download_alter(df_download, fields=self._get_fields_info())
            return df_local
        else:
            return df_download

    def download_resource_bytes(self, ckan:CkanApi, full_download:bool=True, **kwargs) -> bytes:
        df = self.download_resource_df(ckan=ckan, search_all=full_download, **kwargs)
        return self.local_file_format.write_in_memory(df, fields=self._get_fields_info())

    def download_sample_df(self, ckan:CkanApi, *, limit:int=100,
                           search_all:bool=False, download_alter:bool=False, pop_id:bool=True,
                           **kwargs) -> Union[pd.DataFrame,None]:
        """
        Download the first lines of a DataStore. Extra options apply to datastore_dump API.
        """
        df = self.download_resource_df(ckan=ckan, limit=limit, search_all=search_all, download_alter=download_alter, **kwargs)
        if pop_id:
            df.pop(datastore_id_col)
        return df

class BuilderResourceIgnored(BuilderDataStoreABC):
    """
    Class to maintain a line in the resource builders list but has no action and can hold field metadata.
    """
    def __init__(self, *, parent, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, file_url:str=None, options_string:str=None, base_dir:str=None):
        super().__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id,
                         download_url=download_url, options_string=options_string, base_dir=base_dir)
        self.file_url: Union[str, None] = file_url

    def copy(self, *, dest=None, parent=None):
        if dest is None:
            dest = BuilderResourceIgnored(parent=self.parent_package)
        super().copy(dest=dest, parent=parent)
        dest.file_url = self.file_url
        return dest

    @staticmethod
    def resource_mode_str() -> str:
        return "Ignored"

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        self.file_url: str = _string_from_element(row["file/url"], strip=True)
        self._check_mandatory_attributes()

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.file_url
        return d

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return False

    def get_sample_file_path(self, resources_base_dir:str, ckan:Union[CkanApi,None]=None) -> Union[str,None]:
        return None

    def load_sample_data(self, resources_base_dir:str) -> Union[bytes,None]:
        return None

    def load_sample_df(self, resources_base_dir: str, *, upload_alter: bool = True) -> None:
        return None

    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[ContextErrorLevelMessage,None]:
        return None

    def patch_request(self, ckan:CkanApi, package_id:str, *,
                      reupload:bool=None, override_ckan:bool=False, resources_base_dir:str=None,
                      payload:Union[bytes, io.BufferedIOBase]=None) -> None:
        return None

    def download_request(self, ckan: CkanApi, out_dir: str, *, full_download: bool = True, force: bool = False,
                         threads: int = 1, return_data:bool=False) -> Any:
        return None

    def download_resource_bytes(self, ckan: CkanApi, full_download: bool = True, **kwargs) -> bytes:
        return bytes()




#!python3
# -*- coding: utf-8 -*-
"""
Data format policy representation and enforcing
"""
from typing import List, Any, Iterable, Union
from abc import ABC, abstractmethod
import re
import fnmatch
from collections import OrderedDict

from ckanapi_harvesters.policies.data_format_policy_errors import DataPolicyError, ErrorLevel, _policy_msg
from ckanapi_harvesters.policies.data_format_policy_defs import StringMatchMode


class DataPolicyABC(ABC):
    def __init__(self, error_level:ErrorLevel=ErrorLevel.Information):
        self.error_level: ErrorLevel = error_level

    @abstractmethod
    def to_dict(self) -> dict:
        return {"error_level": self.error_level.name}

    @staticmethod
    @abstractmethod
    def from_dict(d:dict):
        raise NotImplementedError()

    @abstractmethod
    def _load_from_dict(self, d:dict):
        self.error_level = ErrorLevel.from_str(d["error_level"])

    @abstractmethod
    def enforce(self, values: Any, *, context:str=None, verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        raise NotImplementedError()


class DataPolicyElementABC(DataPolicyABC, ABC):
    def __init__(self, mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(error_level=error_level)
        self.mandatory:bool = mandatory
        self.match_mode: StringMatchMode = StringMatchMode.Match

    @abstractmethod
    def to_dict(self) -> dict:
        d = OrderedDict([("mandatory", self.mandatory), ("match_mode", self.match_mode.name)])
        d.update(super().to_dict())
        return d

    @staticmethod
    @abstractmethod
    def from_dict(d:dict):
        raise NotImplementedError()

    @abstractmethod
    def _load_from_dict(self, d:dict) -> None:
        super()._load_from_dict(d)
        self.mandatory = d["mandatory"]
        self.match_mode = StringMatchMode.from_str(d["match_mode"])

    def _enforce_unit_string(self, values: Union[str, List[str]], spec: Union[str, Iterable[str]], *, context:str, verbose:bool,
                             buffer:List[DataPolicyError], add_buffer:bool=True) -> bool:
        if values is None or len(values) == 0:
            return not self.match_mode == StringMatchMode.NotEmpty
        if isinstance(values, str):
            values = [values]
        success = True
        for value in values:
            success_value = True
            if isinstance(spec, str):
                spec = [spec]
            if value is None:
                if self.mandatory:
                    success_value = False
            elif self.match_mode == StringMatchMode.Match:
                success_value = value.lower() in {unit_spec.lower() for unit_spec in spec}
            elif self.match_mode == StringMatchMode.MatchCaseSensitive:
                success_value = value in spec
            elif self.match_mode == StringMatchMode.Regex:
                # TODO: test
                success_value = any([re.match(unit_spec,value,flags=re.IGNORECASE) is not None for unit_spec in spec])
            elif self.match_mode == StringMatchMode.RegexCaseSensitive:
                # TODO: test
                success_value = any([re.match(unit_spec,value) is not None for unit_spec in spec])
            elif self.match_mode == StringMatchMode.Wildcard:
                success_value = any([fnmatch.fnmatch(value, unit_spec) is not None for unit_spec in spec])
            elif self.match_mode == StringMatchMode.WildcardCaseSensitive:
                success_value = any([fnmatch.fnmatchcase(value, unit_spec) is not None for unit_spec in spec])
            if add_buffer and not success_value:
                msg = DataPolicyError(context, self.error_level, f"Value does not match spec '{spec}' ({self.match_mode}): {value}")
                _policy_msg(msg, error_level=self.error_level, buffer=buffer, verbose=verbose)
            success &= success_value
        return success






#!python3
# -*- coding: utf-8 -*-
"""
Methods to load an API key
"""

import os
from warnings import warn
from typing import Dict, Union, Iterable
import getpass
import argparse
import  io

from ckanapi_harvesters.auxiliary.ckan_errors import ApiKeyFileError, HostContraintError
from ckanapi_harvesters.auxiliary.path import sanitize_path, path_rel_to_dir
from ckanapi_harvesters.auxiliary.urls import clean_base_url, url_matches_host, is_valid_url
from ckanapi_harvesters.auxiliary.ckan_defs import environ_keyword



class ApiKey:
    """
    API key storage class.
    """

    def __init__(self, remote_url:str=None, *, apikey:str=None, apikey_file:str=None,
                 api_key_header_name:Union[str, Iterable[str]]=None):
        """
        CKAN Database API key storage class.

        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        """
        if api_key_header_name is None:
            api_key_header_name = "Authorization"
        self._remote_url: str = remote_url
        self._remote_url_constraint: Union[str,None] = None
        self.apikey_file: str = apikey_file  # path to a file containing a valid API key in the first line of text (optional)
        self._apikey: str = apikey  # API key used for restricted package access
        self.api_key_header_name = api_key_header_name

    def __del__(self):
        self.clear()

    def __copy__(self):
        return self.copy()

    @property
    def remote_url(self) -> str:
        return self._remote_url
    @remote_url.setter
    def remote_url(self, url:str) -> None:
        self._remote_url = clean_base_url(url)
        self.apply_constraints()

    @property
    def remote_url_constraint(self) -> Union[str,None]:
        return self._remote_url_constraint
    @remote_url_constraint.setter
    def remote_url_constraint(self, url:Union[str,None]) -> None:
        self._remote_url_constraint = clean_base_url(url)
        self.apply_constraints()

    def apply_constraints(self, *, auto_clear:bool=True, raise_error:bool=False) -> bool:
        if self._remote_url_constraint is not None:
            if self._remote_url is not None:
                if not url_matches_host(host_url=self._remote_url_constraint, url=self._remote_url):
                    msg = HostContraintError(host_url=self._remote_url_constraint, url=self._remote_url)
                    if raise_error:
                        raise msg
                    elif auto_clear:
                        self.clear()
                        warn(str(msg) + "\nClearing API key")
                    else:
                        return False
        return True

    def copy(self, *, dest=None):
        if dest is None:
            dest = ApiKey(self._remote_url)
        dest.apikey_file = self.apikey_file
        dest._apikey = self._apikey
        return dest

    def __str__(self):
        if self._apikey is None:
            return "None"
        elif self._apikey == "":
            return "<empty string>"
        else:
            return "*****"

    @property
    def value(self) -> Union[str,None]:
        return self._apikey
    @value.setter
    def value(self, value:Union[str,None]):
        self._apikey = value

    def is_empty(self):
        return self._apikey is None

    def clear(self) -> None:
        self._apikey = None

    def load_from_environ(self, *, error_not_found:bool=False) -> bool:
        """
        Load CKAN API key from environment variables, by order of priority:

        By default, no environment variables are used.
        """
        return False

    def load_apikey(self, apikey_file:str=None, *, base_dir:str=None, error_not_found:bool=True) -> bool:
        """
        Load the API key from file.
        The file should contain a valid API key in the first line of text.

        :param apikey_file: path to the API key file. The following keywords are accepted:
            - "environ": the API key will be looked up in the environment variable with load_from_environ
        :param base_dir: base directory to find the API key file, if a relative path is provided
        :param error_not_found: option to raise an exception if the API key file is not found
        :return:
        """
        if apikey_file is None:
            apikey_file = self.apikey_file
        apikey_file = path_rel_to_dir(apikey_file, base_dir=base_dir, keyword_exceptions={environ_keyword})
        if apikey_file is None:
            raise ApiKeyFileError('apikey_file is required')
        api_keyword = apikey_file.strip().lower()
        if api_keyword == environ_keyword:
            return self.load_from_environ(error_not_found=error_not_found)
        if not(os.path.isfile(apikey_file)) and not error_not_found:
            msg = f"API key file does not exist: {apikey_file}"
            warn(msg)
            return False
        with open(apikey_file, 'r') as f:
            candidate = f.readline().strip()
            if is_valid_url(candidate):
                # when first line is an url, it restricts the usage of the API key to the given CKAN URL
                self.remote_url_constraint = candidate
                apikey = f.readline().strip()
            else:
                apikey = candidate
            f.close()
        self.value = apikey
        self.apikey_file = apikey_file
        return True

    def get_auth_header(self) -> Dict[str, str]:
        """
        Returns the correct header with the API key for the requests needing it.
        If no API key was loaded, returns an empty dictionary.
        """
        if self.value is not None:
            apikey_encoded = self.value
            if isinstance(self.api_key_header_name, str):
                return {self.api_key_header_name: apikey_encoded}
            else:
                return {key: apikey_encoded for key in self.api_key_header_name}
        else:
            return {}

    def input(self):
        """
        Prompt the user to input the API key in the console window.

        :return:
        """
        api_key = getpass.getpass("Please enter the API key: ")
        self._apikey = api_key

    @staticmethod
    def _setup_cli_parser(parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="API key initialization", add_help=False)
        parser.add_argument("--apikey", type=str,
                            help="API key")
        parser.add_argument("--apikey-file", type=str,
                            help="Path to a file containing the API key (first line)")
        return parser

    def print_help_cli(self, display:bool=True) -> str:
        parser = self._setup_cli_parser()
        if display:
            parser.print_help()
        with io.StringIO() as stream:
            parser.print_help(stream)
            return stream.getvalue()

    def _cli_args_apply(self, args: argparse.Namespace, *, base_dir: str = None, error_not_found: bool = True) -> None:
        if args.apikey is not None:
            self.value = args.apikey
        if args.apikey_file is not None:
            self.load_apikey(args.apikey_file, base_dir=base_dir, error_not_found=error_not_found)


class CkanApiKey(ApiKey):
    """
    CKAN Database API key storage class.
    """
    CKAN_API_KEY_HEADER_NAME = {"Authorization", "X-CKAN-API-Key"}  # match apikey_header_name of your CKAN instance
    CKAN_API_KEY_ENVIRON = "CKAN_API_KEY"  # not recommended to store sensitive information in environment variables
    API_KEY_FILE_ENVIRON = "CKAN_API_KEY_FILE"
    API_KEY_FILE_DEFAULT_LIST = [os.path.expanduser(os.path.join("~", ".config", "__CKAN_API_KEY__.txt")),
                                 os.path.expanduser(os.path.join("~", ".ckan", "__CKAN_API_KEY__.txt"))]  # default API key file locations for CKAN

    def __init__(self, apikey:str=None, *, remote_url:str=None, apikey_file:str=None, apikey_auto_load:bool=True):
        """
        CKAN Database API key storage class.

        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        :param apikey_auto_load: option to automatically load the API key from file or envrionment variables

        Order of priority:
        ====
        1. Value of `apikey`
        2. Contents of file pointed by `apikey_file`
        3. Value of environment variable `CKAN_API_KEY`
        4. Contents of file pointed by the environment variable `CKAN_API_KEY_FILE`
        5. Contents of the file at the default location: `~/.config/__CKAN_API_KEY__.txt` or `~/.ckan/__CKAN_API_KEY__.txt`
        """
        super().__init__(remote_url=remote_url, apikey=apikey, apikey_file=apikey_file, api_key_header_name=self.CKAN_API_KEY_HEADER_NAME)
        if apikey_auto_load:
            if self.is_empty() and apikey_file is not None:
                self.load_apikey(error_not_found=False)
            if self.is_empty():
                if self.load_from_environ(error_not_found=False, empty_warning=False):
                    print(f"API key loaded from file: {self.apikey_file}")

    def copy(self, *, dest=None) -> "CkanApiKey":
        if dest is None:
            dest = CkanApiKey(self._remote_url)
        super().copy(dest=dest)
        return dest

    @staticmethod
    def get_default_apikey_file() -> Union[str, None]:
        env_apikey_file = os.environ.get(CkanApiKey.API_KEY_FILE_ENVIRON)   # "CKAN_API_KEY_FILE"
        if env_apikey_file:
            assert not env_apikey_file.strip().lower() == environ_keyword  # this value would create an infinite loop
            return sanitize_path(env_apikey_file)
        else:
            for path_index, default_path in enumerate(CkanApiKey.API_KEY_FILE_DEFAULT_LIST):
                if os.path.exists(default_path):
                    return default_path
        return None

    def load_from_environ(self, *, error_not_found:bool=False, empty_warning:bool=True) -> bool:
        """
        Load CKAN API key from environment variables, by order of priority:

        - `CKAN_API_KEY`: for the raw API key (it is not recommended to store API key in an environment variable)
        - `CKAN_API_KEY_FILE`: path to a file containing a valid API key in the first line of text

        :param error_not_found: raise an error if the API key file was not found
        :return:
        """
        apikey = os.environ.get(self.CKAN_API_KEY_ENVIRON)            # "CKAN_API_KEY"
        default_apikey_file = CkanApiKey.get_default_apikey_file()
        if apikey is not None:
            msg = f"It is not recommended to store sensitive information in environment variables such as the API key ({self.CKAN_API_KEY_ENVIRON})"
            warn(msg)
            self.value = apikey
            return True
        elif default_apikey_file is not None:
            return self.load_apikey(default_apikey_file, error_not_found=error_not_found)
        else:
            if empty_warning:
                msg = f"No API key was found in the environment variable {self.CKAN_API_KEY_ENVIRON}"
                warn(msg)
            return False

    def input(self):
        """
        Prompt the user to input the API key in the console window.

        :return:
        """
        api_key = getpass.getpass("Please enter the CKAN API key: ")
        self._apikey = api_key

    @staticmethod
    def _setup_cli_parser(parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="CKAN API key initialization", add_help=False)
        ApiKey._setup_cli_parser(parser=parser)
        return parser




#!python3
# -*- coding: utf-8 -*-
"""

"""
from typing import List, Tuple, Generator, Any, Union, OrderedDict
import io
import json
from warnings import warn

import numpy as np
import requests
from requests.auth import AuthBase
import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.auxiliary.list_records import ListRecords, records_to_df
from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo, CkanAliasInfo
from ckanapi_harvesters.auxiliary.ckan_map import CkanMap
from ckanapi_harvesters.auxiliary.ckan_auxiliary import bytes_to_megabytes
from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise, CkanIdFieldTreatment
from ckanapi_harvesters.auxiliary.ckan_auxiliary import datastore_id_col
from ckanapi_harvesters.auxiliary.ckan_auxiliary import RequestType
from ckanapi_harvesters.auxiliary.ckan_action import CkanActionResponse, CkanNotFoundError, CkanSqlCapabilityError
from ckanapi_harvesters.auxiliary.ckan_errors import (IntegrityError, CkanServerError, CkanArgumentError, SearchAllNoCountsError,
                                                      DataStoreNotFoundError, RequestError)
from ckanapi_harvesters.auxiliary.ckan_progress_callbacks import CkanProgressCallbackABC
from ckanapi_harvesters.ckan_api.ckan_api_params import CkanApiParamsBasic
from ckanapi_harvesters.auxiliary.ckan_api_key import CkanApiKey
from ckanapi_harvesters.ckan_api.ckan_api_0_base import ckan_request_proxy_default_auth_if_ckan

from ckanapi_harvesters.ckan_api.ckan_api_1_map import CkanApiMap

df_download_read_csv_kwargs = dict(keep_default_na=False)

ckan_dtype_mapper = {
    "text": "str",
    "numeric": "float",
    "timestamp": "datetime64",
    "int": "int",
    "name": "str",
    "oid": "str",  # to confirm
    "bool": "object",  # enable None values but if they are present, booleans are converted to str...
    "json": "object",
}

class CkanApiReadOnlyParams(CkanApiParamsBasic):
    map_all_aliases:bool = True
    default_df_download_id_field_treatment: CkanIdFieldTreatment = CkanIdFieldTreatment.SetIndex

    def __init__(self, *, proxies:Union[str,dict,ProxyConfig]=None,
                 ckan_headers:dict=None, http_headers:dict=None):
        super().__init__(proxies=proxies, ckan_headers=ckan_headers, http_headers=http_headers)
        self.df_download_id_field_treatment: CkanIdFieldTreatment = self.default_df_download_id_field_treatment

    def copy(self, new_identifier:str=None, *, dest=None):
        if dest is None:
            dest = CkanApiReadOnlyParams()
        super().copy(dest=dest)
        dest.df_download_id_field_treatment = self.df_download_id_field_treatment
        return dest


## Main class ------------------
class CkanApiReadOnly(CkanApiMap):
    """
    CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.
    This class implements requests to read data from the CKAN server resources / DataStores.
    """

    def __init__(self, url:str=None, *, proxies:Union[str,dict,ProxyConfig]=None,
                 apikey:Union[str,CkanApiKey]=None, apikey_file:str=None,
                 owner_org:str=None, params:CkanApiReadOnlyParams=None,
                 map:CkanMap=None,
                 identifier=None):
        """
        CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.

        :param url: url of the CKAN server
        :param proxies: proxies to use for requests
        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        :param owner_org: name of the organization to limit package_search (optional)
        :param params: other connection/behavior parameters
        :param map: map of known resources
        :param identifier: identifier of the ckan client
        """
        super().__init__(url=url, proxies=proxies, apikey=apikey, apikey_file=apikey_file,
                         owner_org=owner_org, map=map, identifier=identifier)
        if params is None:
            params = CkanApiReadOnlyParams()
        if proxies is not None:
            params.proxies = proxies
        self.params: CkanApiReadOnlyParams = params

    def _rx_records_df_clean(self, df: pd.DataFrame) -> None:
        """
        Auxiliary function for cleaning dataframe from DataStore requests

        :param df:
        :return:
        """
        if len(df) > 0 and datastore_id_col in df.columns:
            if self.params.df_download_id_field_treatment == CkanIdFieldTreatment.SetIndex:
                # use _id column as new index
                df.set_index(datastore_id_col, drop=False, inplace=True)
                assert(df.index.is_unique)  # verify integrity
            elif self.params.df_download_id_field_treatment == CkanIdFieldTreatment.Remove:
                # remove "_id" column
                df.pop(datastore_id_col)

    @staticmethod
    def read_fields_type_dict(fields_list_dict: List[dict]) -> OrderedDict:
        return OrderedDict([(field_dict["id"], field_dict["type"]) for field_dict in fields_list_dict])

    @staticmethod
    def read_fields_df_args(fields_type_dict: OrderedDict) -> dict:
        if fields_type_dict is None:
            return {}
        # fields_dtype_dict = fields_type_dict.copy()
        # for key, ckan_type in fields_type_dict.items():
        #     if ckan_type in ckan_dtype_mapper:
        #         fields_dtype_dict[key] = ckan_dtype_mapper[ckan_type]
        #     else:
        #         fields_dtype_dict[key] = "object"
        # return dict(names=list(fields_dtype_dict.keys()), dtype=fields_dtype_dict)
        return dict(names=list(fields_type_dict.keys()))

    @staticmethod
    def from_dict_df_args(fields_type_dict: OrderedDict) -> dict:
        df_args_dict = CkanApiReadOnly.read_fields_df_args(fields_type_dict)
        df_args_dict.pop("names")
        return df_args_dict

    @staticmethod
    def _get_default_bom_option(bom:bool=None, format:str=None, search_method:bool=False) -> Union[bool, None]:
        """
        API datastore_dump includes an option to return the BOM (Byte Order Mark) for requests in CSV/TSV format.
        The BOM helps text-processing tools and applications determine the encoding of the file e.g. to distinguish
        between UTF-8 and UTF-16.

        .. note :: To correctly handle BOM characters in pandas.read_csv, you should specify encoding=utf-8-sig parameter.
            This is taken into account in the decoding function.
        """
        if search_method:
            if bom is not None:
                msg = "Argument bom is not used with API datastore_search and will be ignored"
                warn(msg)
            return None  # argument bom not used with API datastore_search
        elif bom is not None:
            return bom
        elif format is not None:
            return format.lower() in {"csv", "tsv"}  # equivalent to format.lower() not in {"json", "xml"}
        else:
            return True  # default format is CSV

    ## Data queries ------------------
    ### Dump method ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # NB: dump methods are not exposed to the user by default. Only datastore_search and resource_download methods are exposed.
    def _api_datastore_dump_raw(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                sort:str=None, limit:int=None, offset:int=0, format:str=None, bom:bool=None, params:dict=None,
                                compute_len:bool=False) -> requests.Response:
        """
        URL call to datastore/dump URL. Dumps successive lines in the DataStore.

        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param format: The return format in the returned response (default=csv, tsv, json, xml) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :return: raw response
        """
        if compute_len:
            raise SearchAllNoCountsError("datastore_search", f"format={format}")
        if params is None:
            params = {}
        if offset is None:
            offset = 0
        params["offset"] = offset
        if limit is None:
            limit = self.params.default_limit_read
        if limit is not None:
            params["limit"] = limit
        if filters is not None:
            if isinstance(filters, str):
                # not recommended
                params["filters"] = filters
            else:
                params["filters"] = json.dumps(filters)
        if q is not None:
            params["q"] = q
        if fields is not None:
            params["fields"] = fields
        if sort is not None:
            params["sort"] = sort
        if format is not None:
            format = format.lower()
            params["format"] = format
        if bom is None and format is not None:
            bom = CkanApiReadOnly._get_default_bom_option(bom=bom, format=format, search_method=False)
        if bom is not None:
            params["bom"] = bom
        # params["bom"] = True  # useful?
        response = self._ckan_url_request(f"datastore/dump/{resource_id}", method=RequestType.Get, params=params)
        if response.status_code == 200:
            return response
        elif response.status_code == 404 and "DataStore resource not found" in response.text:
            raise DataStoreNotFoundError(resource_id, response.content.decode())
        else:
            raise CkanServerError(self, response, response.content.decode())

    def _api_datastore_dump_df(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                               sort:str=None, limit:int=None, offset:int=0, format:str=None, bom:bool=None, params:dict=None) -> pd.DataFrame:
        """
        Convert output of _api_datastore_dump_raw to pandas DataFrame.
        """
        response = self._api_datastore_dump_raw(resource_id=resource_id, filters=filters, q=q, fields=fields,
                                                sort=sort, limit=limit, offset=offset, format=format, bom=bom,
                                                params=params, compute_len=False)
        if format is not None:
            format = format.lower()
        bom = CkanApiReadOnly._get_default_bom_option(bom=bom, format=format, search_method=False)
        assert_or_raise(bom is not None, RuntimeError())
        with io.StringIO(response.content.decode()) as stream:
            read_csv_argument_bom = dict(encoding="utf-8-sig") if bom else {}  # dict(encoding="utf-8")
            if format is None or format == "csv":
                response_df = pd.read_csv(stream, **df_download_read_csv_kwargs, **read_csv_argument_bom)
            elif format == "tsv":
                response_df = pd.read_csv(stream, sep="\t", **df_download_read_csv_kwargs, **read_csv_argument_bom)  # not tested
            elif format == "json":
                response_dict = json.load(stream)
                fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response_dict["fields"])
                df_args = CkanApiReadOnly.read_fields_df_args(fields_type_dict)
                response_df = records_to_df(response_dict["records"], df_args)
                response_df.attrs["fields"] = fields_type_dict
            elif format == "xml":
                response_df = pd.read_xml(stream, parser="etree") # , xpath=".//row")  # partially tested  # otherwise, necessitates the installation of parser lxml
            else:
                raise NotImplementedError()
        self._rx_records_df_clean(response_df)
        return response_df

    def _api_datastore_dump_all(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                sort:str=None, limit:int=None, offset:int=0, format:str=None, bom:bool=None,
                                requests_limit:int=None, params:dict=None, search_all:bool=True, return_df:bool=True,
                                progress_callback:CkanProgressCallbackABC=None) \
            -> Union[pd.DataFrame, requests.Response]:
        """
        Successive calls to _api_datastore_dump_df until an empty list is received.

        :see: _api_datastore_dump()
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param format: The return format in the returned response (default=csv, tsv, json, xml) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            return self._request_all_results_df(api_fun=self._api_datastore_dump_df, params=params,
                                                limit=limit, offset=offset,
                                                requests_limit=requests_limit, progress_callback=progress_callback,
                                                search_all=search_all, resource_id=resource_id,
                                                filters=filters, q=q, fields=fields, sort=sort, format=format, bom=bom)
        elif search_all:
            # cannot determine the number of records received if the response is not parsed with pandas in this mode
            # at least, the total number of rows should be known
            # concatenation of results requires parsing of the result
            # => this mode is useless => raise error
            raise SearchAllNoCountsError("datastore_dump")
        else:
            response = self._api_datastore_dump_raw(resource_id=resource_id, filters=filters, q=q, fields=fields,
                                                    sort=sort, limit=limit, offset=offset, format=format, bom=bom,
                                                    params=params, compute_len=search_all)
            return response

    def _api_datastore_dump_all_page_generator(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                               sort:str=None, limit:int=None, offset:int=0,
                                               requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None,
                                               format:str=None, bom:bool=None,
                                               params:dict=None, search_all:bool=True, return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[requests.Response, Any, None]]:
        """
        Successive calls to _api_datastore_dump until an empty list is received.
        Generator implementation which yields one DataFrame per request.

        :see: _api_datastore_dump()
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param format: The return format in the returned response (default=csv, tsv, json, xml) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            return self._request_all_results_page_generator(api_fun=self._api_datastore_dump_df, params=params,
                                                            limit=limit, offset=offset,
                                                            requests_limit=requests_limit, progress_callback=progress_callback,
                                                            search_all=search_all, resource_id=resource_id,
                                                            filters=filters, q=q, fields=fields, sort=sort, format=format, bom=bom)
        else:
            return self._request_all_results_page_generator(api_fun=self._api_datastore_dump_raw, params=params,
                                                            limit=limit, offset=offset,
                                                            requests_limit=requests_limit, progress_callback=progress_callback,
                                                            search_all=search_all, resource_id=resource_id,
                                                            filters=filters, q=q, fields=fields, sort=sort, format=format, bom=bom,
                                                            compute_len=search_all)


    ### Search method ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    def _api_datastore_search_raw(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                  distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, format:str=None,
                                  params:dict=None, compute_len:bool=False) -> CkanActionResponse:
        """
        API call to datastore_search. Performs queries on the DataStore.

        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param format: The return format in the returned response (default=objects, csv, tsv, lists) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :return:
        """
        if params is None:
            params = {}
        if offset is None:
            offset = 0
        params["offset"] = offset
        if limit is None:
            limit = self.params.default_limit_read
        if limit is not None:
            params["limit"] = limit
        params["resource_id"] = resource_id
        if filters is not None:
            if isinstance(filters, str):
                # not recommended
                params["filters"] = filters
            else:
                params["filters"] = json.dumps(filters)
        if q is not None:
            params["q"] = q
        if fields is not None:
            params["fields"] = fields
        if distinct is not None:
            params["distinct"] = distinct
        if sort is not None:
            params["sort"] = sort
        if format is not None:
            format = format.lower()
            params["records_format"] = format
        response = self._api_action_request(f"datastore_search", method=RequestType.Get, params=params)
        if response.success:
            if response.dry_run:
                return response
            elif format is None or format in ["objects", "lists"]:
                response.len = len(response.result["records"])
                response.total_len = response.result["total"]
            elif compute_len:
                raise SearchAllNoCountsError("datastore_search", f"format={format}")
            return response
        elif response.status_code == 404 and response.error_message["__type"] == "Not Found Error":
            raise DataStoreNotFoundError(resource_id, response.error_message)
        else:
            raise response.default_error(self)

    def _api_datastore_search_df(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                 distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, format:str=None,
                                 params:dict=None, compute_len:bool=True) -> pd.DataFrame:
        """
        Convert output of _api_datastore_search_raw to pandas DataFrame.
        """
        response = self._api_datastore_search_raw(resource_id=resource_id, filters=filters, q=q, fields=fields, format=format,
                                                  distinct=distinct, sort=sort, limit=limit, offset=offset,
                                                  params=params, compute_len=False)
        if response.dry_run:
            return pd.DataFrame()
        if format is not None:
            format = format.lower()
        fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
        if format is None or format == "objects":
            df_args_dict = CkanApiReadOnly.from_dict_df_args(fields_type_dict)
            response_df = pd.DataFrame.from_dict(response.result["records"], **df_args_dict)
        else:
            df_args = CkanApiReadOnly.read_fields_df_args(fields_type_dict)
            if format == "lists":
                response_df = records_to_df(response.result["records"], df_args)
            else:
                with io.StringIO(response.result["records"]) as stream:
                    if format == "csv":
                        response_df = pd.read_csv(stream, **df_args, **df_download_read_csv_kwargs)
                    elif format == "tsv":
                        response_df = pd.read_csv(stream, sep='\t', **df_args, **df_download_read_csv_kwargs)
                    else:
                        raise NotImplementedError()
        self._rx_records_df_clean(response_df)
        response.result.pop("records")
        response_df.attrs["result"] = response.result
        response_df.attrs["fields"] = fields_type_dict
        response_df.attrs["total"] = response.result["total"]
        response_df.attrs["total_was_estimated"] = response.result["total_was_estimated"]
        response_df.attrs["limit"] = response.result["limit"]
        return response_df

    def _api_datastore_search_all(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                  distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, format:str=None,
                                  requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None,
                                  search_all:bool=True, params:dict=None, return_df:bool=True, compute_len:bool=False) \
            -> Union[pd.DataFrame, ListRecords, Any]:
        """
        Successive calls to _api_datastore_search_df until an empty list is received.

        :see: _api_datastore_search()
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param format: The return format in the returned response (default=objects, csv, tsv, lists) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            df = self._request_all_results_df(api_fun=self._api_datastore_search_df, params=params, limit=limit, offset=offset,
                                              requests_limit=requests_limit, progress_callback=progress_callback,
                                              search_all=search_all, resource_id=resource_id,
                                              filters=filters, q=q, fields=fields, distinct=distinct, sort=sort, format=format)
            if "fields" in df.attrs.keys():
                df.attrs["fields"] = df.attrs["fields"][0]
            if "total" in df.attrs.keys():
                assert_or_raise(np.all(np.array(df.attrs["total"]) == df.attrs["total"][0]), IntegrityError("total field varied in the responses"))
                df.attrs["total"] = df.attrs["total"][0]
            return df
        else:
            responses = self._request_all_results_list(api_fun=self._api_datastore_search_raw, params=params, limit=limit, offset=offset,
                                            requests_limit=requests_limit, progress_callback=progress_callback,
                                            search_all=search_all, resource_id=resource_id,
                                            filters=filters, q=q, fields=fields, distinct=distinct, sort=sort, format=format, compute_len=compute_len)
            # aggregate results, depending on the format
            if self.params.dry_run:
                return []
            if format is not None:
                format = format.lower()
            if len(responses) > 0:
                response = responses[0]
                fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
                df_args = CkanApiReadOnly.read_fields_df_args(fields_type_dict)
            else:
                fields_type_dict = None
                df_args = {}
            if format is None or format == "objects":
                return ListRecords(sum([response.result["records"] for response in responses], []))
            else:
                if format == "lists":
                    return sum([response.result["records"] for response in responses], [])
                else:
                    return "\n".join([response.result["records"] for response in responses])

    def _api_datastore_search_all_page_generator(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                                 distinct:bool=None, sort:str=None, limit:int=None, offset:int=0,
                                                 format:str=None, search_all:bool=True,
                                                 requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None,
                                                 params:dict=None, return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[CkanActionResponse, Any, None]]:
        """
        Successive calls to _api_datastore_search_df until an empty list is received.
        Generator implementation which yields one DataFrame per request.

        :see: _api_datastore_search()
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param format: The return format in the returned response (default=objects, csv, tsv, lists) (optional)
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            return self._request_all_results_page_generator(api_fun=self._api_datastore_search_df, params=params, limit=limit, offset=offset,
                                                            requests_limit=requests_limit, progress_callback=progress_callback,
                                                            search_all=search_all, resource_id=resource_id,
                                                            filters=filters, q=q, fields=fields, distinct=distinct, sort=sort, format=format, compute_len=True)
        else:
            return self._request_all_results_page_generator(api_fun=self._api_datastore_search_raw, params=params,
                                                            limit=limit, offset=offset, search_all=search_all,
                                                            requests_limit=requests_limit, progress_callback=progress_callback,
                                                            resource_id=resource_id, filters=filters, q=q,
                                                            fields=fields, distinct=distinct, sort=sort,
                                                            format=format, compute_len=search_all)


    ### search_sql method ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    def _api_datastore_search_sql_raw(self, sql:str, *, params:dict=None, limit:int=None, offset:int=0) -> CkanActionResponse:
        """
        API call to datastore_search_sql. Performs SQL queries on the DataStore. These queries can be more complex than
        with datastore_search. The DataStores are referenced by their resource_id, surrounded by quotes. The field names
        are referred by their name in upper case, surrounded by quotes.
        __NB__: This action is not available when ckanapi_harvesters.datastore.sqlsearch.enabled is set to false

        :param sql: SQL query e.g. f'SELECT * IN "{resource_id}" WHERE "USER_ID" < 0'
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param params: N/A
        :return:
        """
        if params is None:
            params = {}
        params["sql"] = sql
        if offset is None:
            offset = 0
        params["offset"] = offset
        if limit is None:
            limit = self.params.default_limit_read
        if limit is not None:
            params["limit"] = limit
        response = self._api_action_request(f"datastore_search_sql", method=RequestType.Post, params=params)
        if response.success:
            if response.dry_run:
                return response
            else:
                response.len = len(response.result["records"])
            return response
        elif response.status_code == 400 and response.success_json_loads and response.response.text == '"Bad request - Action name not known: datastore_search_sql"':
            raise CkanSqlCapabilityError(self, response)
        elif response.status_code == 404 and response.success_json_loads and response.error_message["__type"] == "Not Found Error":
            raise CkanNotFoundError(self, "SQL", response)
        else:
            raise response.default_error(self)

    def _api_datastore_search_sql_df(self, sql:str, *, params:dict=None, limit:int=None, offset:int=0) -> pd.DataFrame:
        """
        Convert output of _api_datastore_search_sql_raw to pandas DataFrame.
        """
        response = self._api_datastore_search_sql_raw(sql=sql, params=params, limit=limit, offset=offset)
        fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
        df_args_dict = CkanApiReadOnly.from_dict_df_args(fields_type_dict)
        response_df = pd.DataFrame.from_dict(response.result["records"], **df_args_dict)
        response.result.pop("records")
        response_df.attrs["result"] = response.result
        response_df.attrs["fields"] = fields_type_dict
        # response_df.attrs["total"] = response.result["total"]
        # response_df.attrs["total_was_estimated"] = response.result["total_was_estimated"]
        response_df.attrs["limit"] = response.result["limit"]
        self._rx_records_df_clean(response_df)
        return response_df

    def _api_datastore_search_sql_all(self, sql:str, *, params:dict=None,
                                      search_all:bool=True, limit:int=None, offset:int=0,
                                      requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None,
                                      return_df:bool=True) \
            -> Union[pd.DataFrame, ListRecords]:
        """
        Successive calls to _api_datastore_search_sql until an empty list is received.

        :see: _api_datastore_search_sql()
        :param sql: SQL query e.g. f'SELECT * IN "{resource_id}" WHERE "USER_ID" < 0'
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param params: N/A
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            df = self._request_all_results_df(api_fun=self._api_datastore_search_sql_df, params=params,
                                              limit=limit, offset=offset,
                                              requests_limit=requests_limit, progress_callback=progress_callback,
                                              search_all=search_all, sql=sql)
            if "fields" in df.attrs.keys():
                df.attrs["fields"] = df.attrs["fields"][0]
            # if "total" in df.attrs.keys():
            #     assert_or_raise(np.all(np.array(df.attrs["total"]) == df.attrs["total"][0]), IntegrityError("total field varied in the responses"))
            #     df.attrs["total"] = df.attrs["total"][0]
            return df
        else:
            responses = self._request_all_results_list(api_fun=self._api_datastore_search_sql_raw, params=params,
                                                       limit=limit, offset=offset, requests_limit=requests_limit,
                                                       progress_callback=progress_callback,
                                                       search_all=search_all, sql=sql)
            # TODO: test
            if len(responses) > 0:
                response = responses[0]
                fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
            else:
                fields_type_dict = None
            return ListRecords(sum([response.result["records"] for response in responses], []))

    def _api_datastore_search_sql_all_page_generator(self, sql:str, *, params:dict=None,
                                                     search_all:bool=True, limit:int=None, offset:int=0,
                                                     requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None,
                                                     return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[CkanActionResponse, Any, None]]:
        """
        Successive calls to _api_datastore_search_sql until an empty list is received.
        Generator implementation which yields one DataFrame per request.

        :see: _api_datastore_search_sql()
        :param sql: SQL query e.g. f'SELECT * IN "{resource_id}" WHERE "USER_ID" < 0'
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param params: N/A
        :param search_all: if False, only the first request is operated
        :return:
        """
        if return_df:
            return self._request_all_results_page_generator(api_fun=self._api_datastore_search_sql_df, params=params,
                                                            limit=limit, offset=offset, search_all=search_all,
                                                            requests_limit=requests_limit, progress_callback=progress_callback,
                                                            sql=sql)
        else:
            return self._request_all_results_page_generator(api_fun=self._api_datastore_search_sql_raw, params=params,
                                                            limit=limit, offset=offset, search_all=search_all,
                                                            requests_limit=requests_limit, progress_callback=progress_callback,
                                                            sql=sql)


    ## Function aliases to limit the entry-points for the user  -------------------------------------------------------
    def datastore_search(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                         distinct:bool=None, sort:str=None, limit:int=None, offset:int=0, params:dict=None,
                         search_all:bool=False, search_method:bool=True,
                         requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None,
                         format:str=None, bom:bool=None, return_df:bool=True) \
            -> Union[pd.DataFrame, ListRecords, Any, List[CkanActionResponse]]:
        """
        Preferred entry-point for a DataStore read request.
        Uses the API datastore_search

        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param requests_limit: Limit the number of requests
        :param progress_callback: Progress callback function
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: Option to renew the request until there are no more records.
        :param search_method: API method selection (True=datastore_search, False=datastore_dump)
        :return:
        """
        if search_method:
            if return_df and format is None: format = "csv"
            return self._api_datastore_search_all(resource_id, filters=filters, q=q, fields=fields, distinct=distinct, sort=sort,
                                                  limit=limit, offset=offset,
                                                  requests_limit=requests_limit, progress_callback=progress_callback,
                                                  format=format, params=params, search_all=search_all, return_df=return_df)
        else:
            assert_or_raise(distinct is None, CkanArgumentError("DataStore dump", "distinct"))
            if return_df and format is None: format = "csv"
            bom = CkanApiReadOnly._get_default_bom_option(bom=bom, format=format, search_method=search_method)
            return self._api_datastore_dump_all(resource_id, filters=filters, q=q, fields=fields, sort=sort,
                                                limit=limit, offset=offset,
                                                requests_limit=requests_limit, progress_callback=progress_callback,
                                                format=format, bom=bom, params=params, search_all=search_all, return_df=return_df)

    def datastore_dump(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                       distinct:bool=None, sort:str=None, limit:int=None, offset:int=0,
                       requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None,
                       params:dict=None, search_all:bool=True, search_method:bool=True, format:str=None,
                       return_df:bool=True) \
            -> Union[pd.DataFrame, ListRecords, Any, List[CkanActionResponse]]:
        """
        Alias of datastore_search with search_all=True by default.
        Uses the API datastore_search

        :see: datastore_search()
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param requests_limit: Limit the number of requests
        :param progress_callback: Progress callback function
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: Option to renew the request until there are no more records.
        :param search_method: API method selection (True=datastore_search, False=datastore_dump)
        :param return_df: Return pandas DataFrame (True) or dict (False)
        :return:
        """
        return self.datastore_search(resource_id, filters=filters, q=q, fields=fields,
                                     distinct=distinct, sort=sort, limit=limit, offset=offset, requests_limit=requests_limit,
                                     progress_callback=progress_callback, params=params, search_all=search_all,
                                     search_method=search_method, format=format,
                                     return_df=return_df)

    def datastore_search_page_generator(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                        distinct:bool=None, sort:str=None, limit:int=None, offset:int=0,
                                        requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None, params:dict=None,
                                        search_all:bool=True, search_method:bool=True, format:str=None, bom:bool=None, return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[CkanActionResponse, Any, None], Generator[requests.Response, Any, None]]:
        """
        Preferred entry-point for a DataStore read request.
        Uses the API datastore_search

        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param requests_limit: Limit the number of requests
        :param progress_callback: Progress callback function
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: Option to renew the request until there are no more records.
        :param search_method: API method selection (True=datastore_search, False=datastore_dump)
        :param return_df: Return pandas DataFrame (True) or dict (False)
        :return:
        """
        if search_method:
            if return_df and format is None: format = "csv"
            return self._api_datastore_search_all_page_generator(resource_id, filters=filters, q=q, fields=fields, distinct=distinct, sort=sort,
                                                                 limit=limit, offset=offset, requests_limit=requests_limit, progress_callback=progress_callback,
                                                                 format=format, params=params, search_all=search_all, return_df=return_df)
        else:
            assert_or_raise(distinct is None, CkanArgumentError("DataStore dump", "distinct"))
            if return_df and format is None: format = "csv"
            bom = CkanApiReadOnly._get_default_bom_option(bom=bom, format=format, search_method=search_method)
            return self._api_datastore_dump_all_page_generator(resource_id, filters=filters, q=q, fields=fields, sort=sort,
                                                               limit=limit, offset=offset, requests_limit=requests_limit, progress_callback=progress_callback,
                                                               format=format, bom=bom, params=params, search_all=search_all, return_df=return_df)

    def datastore_search_cursor(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                distinct:bool=None, sort:str=None, limit:int=None, offset:int=0,
                                total_limit:int=None, requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None, params:dict=None,
                                search_all:bool=True, search_method:bool=True, format:str=None, bom:bool=None, return_df:bool=False) \
            -> Generator[Union[pd.Series, dict, list, str], Any, None]:
        """
        Cursor on rows of datastore_search

        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param total_limit: Limit the number of records to return
        :param requests_limit: Limit the number of requests
        :param progress_callback: Progress callback function
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: Option to renew the request until there are no more records.
        :param search_method: API method selection (True=datastore_search, False=datastore_dump)
        :param return_df: Return pandas Series (True) or dict (False)
        :param format: Format of the data requested through the API. This does not change the output if return_df is True.
        :return:
        """
        if total_limit is not None and total_limit <= 0:
            return
        row_index = 0
        page_generator = self.datastore_search_page_generator(resource_id, filters=filters, q=q, fields=fields,
                                                              distinct=distinct, sort=sort, limit=limit, offset=offset,
                                                              requests_limit=requests_limit, progress_callback=progress_callback, params=params,
                                                              search_all=search_all, search_method=search_method, format=format, bom=bom, return_df=return_df)
        if return_df:
            df: pd.DataFrame
            row: pd.Series
            for df in page_generator:
                for row_loc, row in df.iterrows():
                    yield row
                    row_index += 1
                    if total_limit is not None and row_index >= total_limit:
                        return
        elif search_method:
            response: CkanActionResponse
            # response.result: list
            if format is not None:
                format = format.lower()
            if format is None or format == "objects":
                for response in page_generator:
                    fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
                    for element in response.result["records"]:
                        yield element
                        row_index += 1
                        if total_limit is not None and row_index >= total_limit:
                            return
            else:
                for response in page_generator:
                    fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
                    for element in response.result["records"]:
                        yield element
                        row_index += 1
                        if total_limit is not None and row_index >= total_limit:
                            return
        else:
            raise TypeError("dumping datastore without parsing with a DataFrame does not return an iterable object")

    def datastore_dump_page_generator(self, resource_id:str, *, filters:dict=None, q:str=None, fields:List[str]=None,
                                      distinct:bool=None, sort:str=None, limit:int=None, offset:int=0,
                                      requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None, params:dict=None,
                                      search_all:bool=True, search_method:bool=True, format:str=None, bom:bool=None, return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[CkanActionResponse, Any, None]]:
        """
        Function alias to datastore_search_generator with search_all=True by default.
        Uses the API datastore_search

        :see: datastore_search_generator
        :param resource_id: resource id.
        :param filters: The base argument to filter values in a table (optional)
        :param q: Full text query (optional)
        :param fields: The base argument to filter columns (optional)
        :param distinct: return only distinct rows (optional, default: false) e.g. to return distinct ids: fields="id", distinct=True
        :param sort: Argument to sort results e.g. sort="index, quantity desc"  or  sort="index asc"
        :param limit: Limit the number of records per request
        :param requests_limit: Limit the number of requests
        :param progress_callback: Progress callback function
        :param offset: Offset in the returned records
        :param params: Additional parameters such as filters, q, sort and fields can be given. See DataStore API documentation.
        :param search_all: Option to renew the request until there are no more records.
        :param search_method: API method selection (True=datastore_search, False=datastore_dump)
        :return:
        """
        return self.datastore_search_page_generator(resource_id, filters=filters, q=q, fields=fields,
                                                    distinct=distinct, sort=sort, limit=limit, offset=offset,
                                                    requests_limit=requests_limit, progress_callback=progress_callback, params=params,
                                                    search_all=search_all, search_method=search_method, format=format, bom=bom, return_df=return_df)

    def datastore_search_sql(self, sql:str, *, params:dict=None, search_all:bool=False,
                             limit:int=None, offset:int=0, requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None, return_df:bool=True) \
            -> Union[pd.DataFrame, Tuple[ListRecords, dict]]:
        """
        Preferred entry-point for a DataStore SQL request.
        :see: _api_datastore_search_sql()
        __NB__: This action is not available when ckanapi_harvesters.datastore.sqlsearch.enabled is set to false

        :param sql: SQL query e.g. f'SELECT * IN "{resource_id}" WHERE "USER_ID" < 0'
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param requests_limit: Limit the number of requests
        :param progress_callback: Progress callback function
        :param params: N/A
        :param search_all: Option to renew the request until there are no more records.
        :param return_df: Return pandas DataFrame (True) or dict (False)
        :return:
        """
        return self._api_datastore_search_sql_all(sql, params=params, limit=limit, offset=offset,
                                                  requests_limit=requests_limit, progress_callback=progress_callback,
                                                  search_all=search_all, return_df=return_df)

    def datastore_search_sql_page_generator(self, sql:str, *, params:dict=None, search_all:bool=True,
                                            limit:int=None, offset:int=0,
                                            requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None,
                                            return_df:bool=True) \
            -> Union[Generator[pd.DataFrame, Any, None], Generator[CkanActionResponse, Any, None]]:
        """
        Preferred entry-point for a DataStore SQL request.
        :see: _api_datastore_search_sql()

        __NB__: This action is not available when ckanapi_harvesters.datastore.sqlsearch.enabled is set to false

        :param sql: SQL query e.g. f'SELECT * IN "{resource_id}" WHERE "USER_ID" < 0'
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param requests_limit: Limit the number of requests
        :param progress_callback: Progress callback function
        :param params: N/A
        :param search_all: Option to renew the request until there are no more records.
        :param return_df: Return pandas DataFrame (True) or dict (False)
        :return:
        """
        return self._api_datastore_search_sql_all_page_generator(sql, params=params, limit=limit, offset=offset,
                                                                 requests_limit=requests_limit, progress_callback=progress_callback,
                                                                 search_all=search_all, return_df=return_df)

    def datastore_search_sql_cursor(self, sql:str, *, params:dict=None, search_all:bool=True,
                                    limit:int=None, offset:int=0, total_limit:int=None,
                                    requests_limit:int=None, progress_callback:CkanProgressCallbackABC=None,
                                    return_df:bool=False) \
            -> Generator[Union[pd.Series,dict], Any, None]:
        """
        Preferred entry-point for a DataStore SQL request, to iterate over records.
        :see: _api_datastore_search_sql()

        __NB__: This action is not available when ckanapi_harvesters.datastore.sqlsearch.enabled is set to false

        :param sql: SQL query e.g. f'SELECT * IN "{resource_id}" WHERE "USER_ID" < 0'
        :param limit: Limit the number of records per request
        :param offset: Offset in the returned records
        :param total_limit: Limit the number of records to return
        :param requests_limit: Limit the number of requests
        :param progress_callback: Progress callback function
        :param params: N/A
        :param search_all: Option to renew the request until there are no more records.
        :param return_df: Return pandas Series (True) or dict (False)
        :return:
        """
        if total_limit is not None and total_limit <= 0:
            return
        row_index = 0
        page_generator = self.datastore_search_sql_page_generator(sql, params=params, search_all=search_all,
                                                                  limit=limit, offset=offset,
                                                                  requests_limit=requests_limit, progress_callback=progress_callback,
                                                                  return_df=return_df)
        if return_df:
            df: pd.DataFrame
            row: pd.Series
            for df in page_generator:
                for row_loc, row in df.iterrows():
                    yield row
                    row_index += 1
                    if total_limit is not None and row_index >= total_limit:
                        return
        else:
            response: CkanActionResponse
            # response.result: list
            element: Any
            for response in page_generator:
                fields_type_dict = CkanApiReadOnly.read_fields_type_dict(response.result["fields"])
                for element in response.result["records"]:
                    yield element
                    row_index += 1
                    if total_limit is not None and row_index >= total_limit:
                        return

    def datastore_search_sql_find_one(self, sql:str, *, params:dict=None,
                                      offset:int=0, return_df:bool=True) -> Union[pd.DataFrame, Tuple[ListRecords, dict]]:
        """
        First element of an SQL request
        """
        df_row = self.datastore_search_sql(sql, limit=1, search_all=False, offset=offset, params=params, return_df=return_df)
        return df_row

    def datastore_search_sql_fields_type_dict(self, sql:str, *, params:dict=None) -> OrderedDict:
        document, fields_dict = self.datastore_search_sql_find_one(sql, offset=0, params=params, return_df=False)
        return fields_dict

    def datastore_search_sql_row_count(self, sql:str, *, params:dict=None) -> int:
        df_row = self.datastore_search_sql_find_one(sql, offset=0, params=params, return_df=True)
        return df_row.attrs["total"]

    def datastore_search_find_one(self, resource_id:str, *, filters:dict=None, q:str=None, distinct:bool=None,
                                  fields:List[str]=None, offset:int=0, return_df:bool=True) \
            -> Union[pd.DataFrame, ListRecords, Any, List[CkanActionResponse]]:
        """
        Request first result of a query

        :param resource_id: resource id
        :return:
        """
        # resource_info = self.get_resource_info_or_request(resource_id)
        # return resource_info.datastore_info.row_count
        df_row = self.datastore_search(resource_id, limit=1, search_all=False, filters=filters, q=q, distinct=distinct,
                                       fields=fields, offset=offset, return_df=return_df)
        return df_row

    def datastore_search_fields_type_dict(self, resource_id:str, *,
                                          filters:dict=None, q:str=None, distinct:bool=None, fields:List[str]=None,
                                          request_missing:bool=True, error_not_mapped:bool=False,
                                          error_not_found:bool=True) -> OrderedDict:
        if fields is None:
            # if no field restriction was provided, refer to the fields of the DataStore
            fields_list = self.get_datastore_fields_or_request(resource_id, return_list=True,
                                                               request_missing=request_missing,
                                                               error_not_mapped=error_not_mapped,
                                                               error_not_found=error_not_found)
            return CkanApiReadOnly.read_fields_type_dict(fields_list)
        else:
            document, fields_dict = self.datastore_search_find_one(resource_id, filters=filters, q=q, distinct=distinct,
                                                                   fields=fields, return_df=False)
            return fields_dict

    def datastore_search_row_count(self, resource_id:str, *, filters:dict=None, q:str=None, distinct:bool=None,
                                   fields:List[str]=None) -> int:
        """
        Request the number of rows in a DataStore

        :param resource_id: resource id
        :return:
        """
        df_row = self.datastore_search_find_one(resource_id, filters=filters, q=q, distinct=distinct,
                                                fields=fields, return_df=True)
        return df_row.attrs["total"]

    def test_sql_capabilities(self, *, raise_error:bool=False) -> bool:
        """
        Test the availability of the API datastore_search_sql

        :return:
        """
        try:
            self.api_help_show("datastore_search_sql", print_output=False)
            return True
        except CkanNotFoundError:
            if raise_error:
                raise CkanSqlCapabilityError(self, CkanActionResponse(requests.Response()))
            return False


    ## Resource download by direct link (FileStore) -----------------------------------------------
    def resource_download(self, resource_id:str, *, method:str=None,
                          proxies:dict=None, headers:dict=None, auth: Union[AuthBase, Tuple[str,str]]=None,
                          verify:Union[bool,str,None]=None, stream:bool=False) \
            -> Tuple[CkanResourceInfo, Union[requests.Response,None]]:
        """
        Uses the link provided in resource_show to download a resource.

        :param resource_id: resource id
        :return:
        """
        resource_info = self.get_resource_info_or_request(resource_id)
        url = resource_info.download_url
        if len(url) == 0:
            return resource_info, None
        response = self.download_url_proxy(url, method=method, auth_if_ckan=ckan_request_proxy_default_auth_if_ckan,
                                           proxies=proxies, headers=headers, auth=auth, verify=verify, stream=stream)
        content_length = response.headers.get("content-length", None)
        if content_length is not None:
            resource_info.download_size_mb = bytes_to_megabytes(int(content_length))
        return resource_info, response

    def resource_download_test_head(self, resource_id:str, *, raise_error:bool=False,
                                    proxies:dict=None, headers:dict=None, auth: Union[AuthBase, Tuple[str,str]]=None, verify:Union[bool,str,None]=None) \
            -> Union[None,ContextErrorLevelMessage]:
        """
        This sends a HEAD request to the resource download url using the CKAN connexion parameters via resource_download.
        The resource is not downloaded but the headers indicate if the url is valid.

        :return: None if successful
        """
        resource_info = self.get_resource_info_or_request_of_id(resource_id)
        try:
            _, response = self.resource_download(resource_id, method="HEAD", proxies=proxies, headers=headers, auth=auth, verify=verify)
        except Exception as e:
            if raise_error:
                raise e from e
            return ContextErrorLevelMessage(f"Resource from URL {resource_info.name}", ErrorLevel.Error, f"Failed to query download url for resource id {resource_id}: {str(e)}")
        if response.ok and response.status_code == 200:
            return None
        else:
            if raise_error:
                raise RequestError(f"Failed to query download url for resource id {resource_id}: status {response.status_code} {response.reason}")
            return ContextErrorLevelMessage(f"Resource from URL {resource_info.name}", ErrorLevel.Error, f"Failed to query download url for resource id {resource_id}: status {response.status_code} {response.reason}")

    def resource_download_df(self, resource_id:str, *, method:str=None,
                          proxies:dict=None, headers:dict=None, auth: Union[AuthBase, Tuple[str,str]]=None, verify:Union[bool,str,None]=None) \
            -> Tuple[CkanResourceInfo, Union[pd.DataFrame,None]]:
        """
        Uses the link provided in resource_show to download a resource and interprets it as a DataFrame.

        :param resource_id: resource id
        :return:
        """
        resource_info, response = self.resource_download(resource_id, method=method, proxies=proxies, headers=headers, auth=auth, verify=verify)
        if response is None:
            return resource_info, None
        with io.StringIO(response.content.decode()) as stream:
            df = pd.read_csv(stream, **df_download_read_csv_kwargs)
        self._rx_records_df_clean(df)
        return resource_info, df

    def map_file_resource_sizes(self, cancel_if_present:bool=True) -> None:
        for resource_id, resource_info in self.map.resources.items():
            if resource_info.download_url:
                if not (cancel_if_present and resource_info.download_size_mb is not None):
                    try:
                        _, response = self.resource_download(resource_id, method="HEAD")
                        content_length_str = response.headers.get("content-length", None)
                    except Exception as e:
                        msg = f"Failed to query download url for resource id {resource_id}: {str(e)}"
                        warn(msg)
                        content_length_str = None
                    if content_length_str is not None:
                        content_length = int(content_length_str)  # raise error if not found or bad format
                        resource_info.download_size_mb = bytes_to_megabytes(content_length)
                    else:
                        resource_info.download_size_mb = None


    ## Mapping of resource aliases from table
    def list_datastore_aliases(self) -> List[CkanAliasInfo]:
        alias_resource_id = "_table_metadata"  # resource name of table containing CKAN aliases
        alias_list_dict = self.datastore_search(alias_resource_id, search_all=True, return_df=False, format="objects", search_method=True)
        alias_list = [CkanAliasInfo(alias_dict) for alias_dict in alias_list_dict]
        for alias_info in alias_list:
            if alias_info.alias_of is not None:
                self.map.resource_alias_index[alias_info.name] = alias_info.alias_of
        return alias_list

    def map_resources(self, package_list:Union[str, List[str]]=None, *, params:dict=None,
                      datastore_info:bool=None, resource_view_list:bool=None, organization_info:bool=None, license_list:bool=None,
                      only_missing:bool=True, error_not_found:bool=True,
                      owner_org:str=None) -> CkanMap:
        # overload including a call to list all aliases
        if len(self.map.resource_alias_index) == 0 and self.params.map_all_aliases:
            self.list_datastore_aliases()
        map = super().map_resources(package_list=package_list, params=params, datastore_info=datastore_info,
                              resource_view_list=resource_view_list, organization_info=organization_info,
                              license_list=license_list, only_missing=only_missing, error_not_found=error_not_found,
                              owner_org=owner_org)
        return map




#!python3
# -*- coding: utf-8 -*-
"""
This implements functionality to dynamically call functions specified by the user.
This functionality is disabled by default. You must call unlock_external_code_execution to enable external code execution.
__Warning__:
only run code if you trust the source!
"""
from typing import Callable
import os
import re
from warnings import warn
import importlib
import importlib.util

from ckanapi_harvesters.auxiliary.path import path_rel_to_dir



def unlock_external_code_execution(value:bool=True) -> None:
    """
    This function enables external code execution for the PythonUserCode class.

    __Warning__:
    only run code if you trust the source!

    :return:
    """
    PythonUserCode.enable_external_code = value
    if value:
        msg = "External code is enabled. Only run code if you trust the source!"
        warn(msg)


var_name_subs_re = "\\W|^(?=\\d)"

def clean_var_name(variable_name: str) -> str:
    return re.sub(var_name_subs_re,'_', variable_name)


class ExternalUserCodeDisabledException(Exception):
    def __init__(self, function_name:str, source_file:str) -> None:
        super().__init__(f"{function_name} in {source_file} cannot be executed because the external code execution is locked. Use unlock_external_code_execution to unlock. Warning: Only run code if you trust the source!")


class PythonUserCode:
    """
    This class imports an arbitrary Python file as a module and makes it available to the rest of the code.
    This functionality is disabled by default. You must call unlock_external_code_execution to enable external code execution.

    __Warning__:
    only run code if you trust the source!
    """
    enable_external_code = False  # remain False to ensure no custom code is executed from the builder specification

    def __init__(self, python_file:str, base_dir:str=None):
        self.python_file: str = ""
        if python_file is None:
            # only use None argument to initialize copy
            self.python_file = ""
        else:
            self.python_file = path_rel_to_dir(python_file, base_dir=base_dir)
        self.module = None
        self.imported_module_name: str = ""
        if PythonUserCode.enable_external_code and python_file is not None:
            module_base = clean_var_name(os.path.splitext(os.path.split(self.python_file)[1])[0])
            self.imported_module_name = "ckan_builder_aux_funcs__" + module_base
            spec = importlib.util.spec_from_file_location(self.imported_module_name, self.python_file)
            self.module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(self.module)
        elif python_file is not None:  # and not enable_external_code
            # External python script execution is locked
            raise ExternalUserCodeDisabledException("code", os.path.split(self.python_file)[1])

    def __copy__(self):
        return self.copy()

    def copy(self) -> "PythonUserCode":
        # do not execute module twice and do not make copies either
        dest = PythonUserCode(python_file=None, base_dir=None)
        dest.module = self.module
        dest.imported_module_name = self.imported_module_name
        dest.python_file = self.python_file
        return dest

    def function_pointer(self, function_name:str) -> Callable:
        """
        Obtain function pointer for a given name in the loaded Python module.

        :param function_name:
        :return:
        """
        if not PythonUserCode.enable_external_code:
            raise ExternalUserCodeDisabledException(function_name, os.path.split(self.python_file)[1])
        fun = getattr(self.module, function_name)
        assert(isinstance(fun, Callable))
        return fun




#!python3
# -*- coding: utf-8 -*-
"""
Functions to clean data before upload.
"""
from typing import Union, List, Any, Type
from collections import OrderedDict
import copy
import math
import numbers
from warnings import warn
import datetime
import json
import re
import base64

import pandas as pd
try:
    import bson
except ImportError:
    bson = None

from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_timestamp_sep
from ckanapi_harvesters.auxiliary.ckan_errors import IntegrityError
from ckanapi_harvesters.auxiliary.list_records import ListRecords, records_to_df
from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise
from ckanapi_harvesters.auxiliary.ckan_auxiliary import datastore_id_col
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_errors import CleanError, CleanerRequirementError
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC

non_finite_authorized_types = {"numeric", "float4", "float8", "float2"}
real_number_types = non_finite_authorized_types
int_types = {"int", "int4", "int8", "int2"}
no_empty_str_types = real_number_types.union(int_types).union({"bool", "timestamp", "date"})
# see also: ckan_api_2_readonly ckan_dtype_mapper
dtype_ckan_mapper = {
    "float64": "numeric",
    "int64": "numeric",
    "datetime64[ns]": "timestamp",
}


def _pd_series_type_instance_detect(values: pd.Series, test_type:Type):
    """
    This function checks that the test_type matches all rows which are not NaN/None/NA in a pandas Series.
    """
    return values.map(lambda x: isinstance(x, test_type)).where(values.notna(), True).all()

def _pd_str_series_type_detect(values: pd.Series) -> str:
    # remove NA values
    series = values.where(values.notna(), None)
    if len(series) == 0:
        return "text"  # could not determine data type

    # Check for booleans
    if series.str.strip().str.lower().isin(['true', 'false']).all():
        return 'bool'

    # Check for integers
    float_values = pd.to_numeric(series, errors='coerce', downcast="float")
    if float_values.apply(float.is_integer).all():
        return 'int'

    # Check for floats
    if float_values.notna().all():
        return 'float'

    # Check for timestamps
    if pd.to_datetime(series, errors='coerce').notna().all():
        return 'timestamp'

    return 'text'


class CkanDataCleanerUploadBasic(CkanDataCleanerABC):
    """
    Data cleaner for basic data types
    """
    def __init__(self):
        super().__init__()
        # options
        self.param_json_as_text:bool = False  # option to convert json fields (dicts and lists) to str
        self.param_replace_nan:bool = True  # option to replace non-authorized nan values by None
        self.param_round_values:bool = True  # option to round values when treating an integer field
        self.param_rename_fields_underscore:bool = True  # option to rename fields beginning with an underscore (in the subs step)
        self.param_empty_str_digital: bool = True  # option to replace empty strings with None for digital column types (real, int, timestamp, bool)

    @staticmethod
    def get_class_keyword() -> str:
        return "Basic"

    def copy(self, dest=None) -> "CkanDataCleanerUploadBasic":
        if dest is None:
            dest = CkanDataCleanerUploadBasic()
        super().copy(dest=dest)
        dest.param_json_as_text = self.param_json_as_text
        dest.param_replace_nan = self.param_replace_nan
        dest.param_round_values = self.param_round_values
        dest.param_rename_fields_underscore = self.param_rename_fields_underscore
        dest.param_field_subs = self.param_field_subs.copy()
        self.param_empty_str_digital = self.param_empty_str_digital
        return dest

    ## field type detection
    def create_new_field(self, field_name:str, values: Union[Any, pd.Series]) -> CkanField:
        if field_name in self.fields_new.keys():
            return self.fields_new[field_name]
        else:
            # detect type
            if isinstance(values, pd.Series):
                dtype = str(values.dtype)
                if dtype == "object":
                    field_info = self._detect_standard_field_bypass(field_name, values)
                    if field_info is not None:
                        return field_info
                    elif _pd_series_type_instance_detect(values, str):
                        data_type = _pd_str_series_type_detect(values)
                        return CkanField(field_name, data_type)
                    elif _pd_series_type_instance_detect(values, bool):
                        return CkanField(field_name, "bool")
                    elif (_pd_series_type_instance_detect(values, dict)
                          or _pd_series_type_instance_detect(values, list)):
                        if self.param_json_as_text:
                            return CkanField(field_name, "text")
                        else:
                            return CkanField(field_name, "json")
                    elif (_pd_series_type_instance_detect(values, datetime.datetime)
                          or _pd_series_type_instance_detect(values, pd.Timestamp)):
                        return CkanField(field_name, "timestamp")
                    elif _pd_series_type_instance_detect(values, int):
                        return CkanField(field_name, "int")
                    elif _pd_series_type_instance_detect(values, float):
                        return CkanField(field_name, "float")
                    else:
                        return self._detect_non_standard_field(field_name, values)
                elif dtype in dtype_ckan_mapper.keys():
                    return CkanField(field_name, dtype_ckan_mapper[dtype])
                else:
                    return CkanField(field_name, dtype)
            else:
                return CkanField(field_name, str(type(values)))

    def _initial_field_subs(self, fields: OrderedDict[str, CkanField]) -> OrderedDict[str, CkanField]:
        # rename fields beginning with '_'
        for field_name, value in fields.items():
            if field_name not in self.field_subs.keys():
                if field_name in self.param_field_subs.keys():
                    self.field_subs[field_name] = self.param_field_subs[field_name]
                elif self.param_rename_fields_underscore and field_name.startswith("_") and not field_name == datastore_id_col:
                    index = re.search(r"[a-zA-Z]", field_name)
                    if index is not None:
                        self.field_subs[field_name] = field_name[index.start():]
                    else:
                        raise NameError(f"Field {field_name} is invalid")
        return fields

    def detect_field_types_and_subs(self, records: Union[List[dict], pd.DataFrame],
                                    known_fields:OrderedDict[str, CkanField]=None) -> OrderedDict[str, CkanField]:
        self.clear_outputs_new_dataframe()
        fields = OrderedDict()
        if known_fields is not None:
            for field_name, value in known_fields.items():
                fields[field_name] = value
        if isinstance(records, list):
            df = records_to_df(records)
        else:
            df = records
        for column in df.columns:
            if column in self.field_subs.keys():
                column_new = self.field_subs[column]
            else:
                column_new = column
            if known_fields is None or column_new not in known_fields.keys():
                fields[column_new] = self.create_new_field(column_new, df[column])
                self.fields_new[column_new] = fields[column_new]
        fields = self._initial_field_subs(fields)
        return fields

    ## Data cleaning
    def _clean_subvalues_recursive(self, subvalue:Any, field:CkanField, path:str, level:int,
                                   *, field_data_type:str) -> Any:
        if isinstance(subvalue, dict):
            for key, element in subvalue.items():
                if not isinstance(key, str):
                    raise TypeError(f"Key {key} is of invalid type")
                subvalue[key] = self._clean_subvalues_recursive(element, field, path + "." + str(key), level + 1,
                                                                field_data_type=field_data_type)
            return subvalue
        elif isinstance(subvalue, list):
            for i, element in enumerate(subvalue):
                subvalue[i] = self._clean_subvalues_recursive(element, field, path + "[" + str(i) + "]", level + 1,
                                                              field_data_type=field_data_type)
            return subvalue
        else:
            return self._clean_subvalue(subvalue, field, path, level, field_data_type=field_data_type)

    def _clean_subvalue(self, subvalue: Any, field: CkanField, path: str, level: int,
                                   *, field_data_type: str) -> Any:
        field_name = field.name if field is not None else None
        new_subvalue, bypass = self._replace_standard_subvalue_bypass(subvalue, field, path, level, field_data_type=field_data_type)
        if bypass:
            pass  # return new_subvalue
        else:
            new_subvalue = subvalue
            if isinstance(subvalue, numbers.Number):
                if not math.isfinite(subvalue):
                    if math.isnan(subvalue):
                        if self.param_replace_nan:
                            new_subvalue = None  # replace nans with None when not authorized
                        else:
                            self.warnings[field_name].add("nan")
                    else:
                        self.warnings[field_name].add("inf")  # infinite values are not authorized and no replacement can be made
                        if self.param_replace_forbidden:
                            new_subvalue = None
            elif isinstance(subvalue, datetime.datetime):
                if self.param_cast_types:
                    new_subvalue = subvalue.isoformat(sep=ckan_timestamp_sep)
            else:
                new_subvalue = self._replace_non_standard_subvalue(subvalue, field, path, level, field_data_type=field_data_type)
        if path in self.field_subs_path.keys():
            self._new_columns_in_row[path] = new_subvalue
        return new_subvalue

    def clean_value_field(self, value: Any, field:CkanField) -> Any:
        field_name = field.name if field is not None else None
        field_data_type = field.data_type if field is not None else None
        field_data_type = field_data_type.lower() if field_data_type is not None else None
        if field_name not in self.warnings:
            self.warnings[field_name] = set()
            self.fields_encountered[field_name] = None
        new_value, bypass = self._replace_standard_value_bypass(value, field, field_data_type=field_data_type)
        if bypass:
            pass  # return new_value
        else:
            new_value = value
            if isinstance(value, dict) or isinstance(value, list):
                if field_data_type == "text" and self.param_cast_types:
                    return json.dumps(value, default=str)
                elif field_data_type == "bson":
                    if bson is None:
                        raise CleanerRequirementError("bson", "bson")
                    return base64.b64encode(bson.BSON.encode(value))  # TODO: confirm need to encode in base64
                else:
                    return self._clean_subvalues_recursive(subvalue=value, field=field, path=field_name, level=0,
                                                           field_data_type=field_data_type)
            elif isinstance(value, numbers.Number):
                if (not math.isfinite(value)) and field_data_type not in non_finite_authorized_types:
                    if math.isnan(value):
                        if self.param_replace_nan:
                            return None  # replace nans with None when not authorized
                        else:
                            self.warnings[field_name].add("nan")
                    else:
                        self.warnings[field_name].add("inf")  # infinite values are not authorized and no replacement can be made
                        if self.param_replace_forbidden:
                            return None
                elif isinstance(value, bool):
                    if field_data_type == "text":
                        if self.param_cast_types:
                            return str(value)
                    elif field_data_type == "numeric":
                        if self.param_cast_types:
                            return int(value)
                    elif not field_data_type == "bool":
                        self.field_changes[field_name] = CkanField(field_name, "bool")
                elif field_data_type not in real_number_types and not round(value) == value:
                    if self.param_round_values:
                        return round(value)
                    else:
                        self.warnings[field_name].add("float")
            elif isinstance(value, datetime.datetime):
                if field_data_type == "timestamp":
                    if self.param_cast_types:
                        return value.isoformat(sep=ckan_timestamp_sep)
                elif not field_data_type == "timestamp":
                    self.field_changes[field_name] = CkanField(field_name, "timestamp")
                    if self.param_cast_types:
                        return value.isoformat(sep=ckan_timestamp_sep)
            elif isinstance(value, str) and value == "":
                if field_data_type in no_empty_str_types:
                    if self.param_empty_str_digital:
                        new_value = None
            else:
                new_value = self._replace_non_standard_value(value, field, field_data_type=field_data_type)
        return new_value

    def clean_records(self, records: Union[List[dict], pd.DataFrame],
                      known_fields:Union[OrderedDict[str, CkanField], OrderedDict[str,dict], List[Union[dict,CkanField]], None],
                      *, inplace:bool=False) -> Union[List[dict], pd.DataFrame]:
        self.clear_outputs_new_dataframe()
        if known_fields is not None and isinstance(known_fields, list):
            fields_list = known_fields
            known_fields = OrderedDict()
            for field_info in fields_list:
                if isinstance(field_info, dict):
                    field_dict = field_info
                    field_info = CkanField.from_ckan_dict(field_dict)
                known_fields[field_info.name] = field_info
        elif known_fields is not None and isinstance(known_fields, dict):
            for field_name, field_info in known_fields.items():
                if isinstance(field_info, dict):
                    field_dict = field_info
                    field_info = CkanField.from_ckan_dict(field_dict)
                if field_info.name is None:
                    field_info.name = field_name
                else:
                    assert_or_raise(field_info.name == field_name, IntegrityError(f"Field name {field_info.name} neq {field_name}"))
                known_fields[field_info.name] = field_info
        fields = self.detect_field_types_and_subs(records, known_fields=known_fields)
        if not inplace:
            records = copy.deepcopy(records)
        if not self.param_enable:
            return records
        # iterate on records
        mode_df = isinstance(records, pd.DataFrame)
        if mode_df:
            records = records.astype(object)  # Convert all columns to object type
            for new_field in self.field_subs_path.values():
                records[new_field] = None
            for column in records.columns:
                field = fields[column]
                # records[column] = records[column].apply(self.clean_value_field, field=field)
                for value_loc, value in zip(records.index, records[column]):
                    self._new_columns_in_row = {}
                    records.at[value_loc, column] = self.clean_value_field(value, field=field)
                    # copy-columns created at clean_value_field request are stored in attribute _new_columns_in_row
                    for path, new_value in self._new_columns_in_row.items():
                        if path in self.field_subs_path.keys():
                            new_field = self.field_subs_path[path]
                            records.at[value_loc, new_field] = new_value
        else:
            for row in records:
                self._new_columns_in_row = {}
                for key, value in row.items():
                    field = fields[key]
                    row[key] = self.clean_value_field(value, field=field)
                for path, new_value in self._new_columns_in_row.items():
                    if path in self.field_subs_path.keys():
                        new_field = self.field_subs_path[path]
                        assert(new_field not in row.keys())
                        row[new_field] = new_value
                if self.param_apply_field_subs:
                    for field_name, substitution in self.field_subs.items():
                        if field_name in row.keys():
                            assert_or_raise(substitution not in row.keys(), KeyError(substitution))
                            row[substitution] = row.pop(field_name)
        return self._clean_final_steps(records, fields, known_fields)

    def _clean_final_steps(self, records: Union[List[dict], pd.DataFrame], fields:Union[OrderedDict[str, CkanField], None],
                           known_fields:Union[OrderedDict[str, CkanField], None]) -> Union[List[dict], pd.DataFrame]:
        # apply final modifications
        mode_df = isinstance(records, pd.DataFrame)
        self.warnings = {key: value for key, value in self.warnings.items() if len(value) > 0}
        if len(self.warnings) > 0:
            msg = "Some fields had anomalies: " + str(self.warnings)
            if self.param_raise_error:
                raise CleanError(msg)
            elif self.param_verbose:
                warn(msg)
        if len(self.field_subs) > 0:
            for field_name, substitution in self.field_subs.items():
                if substitution in self.fields_encountered.keys():
                    msg = f"Substitution cannot be done for field '{field_name}' because '{substitution}' already exists"
                    if self.param_raise_error or self.param_apply_field_subs:
                        raise KeyError(msg)
                    elif self.param_verbose:
                        warn(msg)
            if self.param_apply_field_subs:
                if mode_df:
                    if len(self.field_subs) > 0:
                        records.rename(columns=self.field_subs, inplace=True)
                    # for field_name, substitution in self.field_subs.items():
                        # records[substitution] = records.pop(field)
                else:
                    pass # already done above
                    # for row in records:
                    #     for field_name, substitution in self.field_subs.items():
                    #         if field_name in row.keys():
                    #             row[substitution] = row.pop(field_name)
                new_encountered_fields = self.fields_encountered
                self.fields_encountered = OrderedDict()
                for field_name in new_encountered_fields.keys():
                    if field_name in self.field_subs.keys():
                        substitution = self.field_subs[field_name]
                        self.fields_encountered[substitution] = None
                    else:
                        self.fields_encountered[field_name] = None
                new_fields_copy = self.fields_new
                self.fields_new = OrderedDict()
                for field_name, field_info in new_fields_copy.items():
                    if field_name in self.field_subs.keys():
                        substitution = self.field_subs[field_name]
                        if known_fields is None or substitution not in known_fields.keys():
                            self.fields_new[substitution] = field_info
                            self.fields_new[substitution].name = substitution
                    elif known_fields is None or field_name not in known_fields.keys():
                        self.fields_new[field_name] = field_info
                    else:
                        pass  # field already known
                for field_name, substitution in self.field_subs.items():
                    if field_name in self.field_changes.keys():
                        self.field_changes[substitution] = self.field_changes.pop(field_name)
                        self.field_changes[substitution].name = substitution
                    if field_name in self.field_suggested_index:
                        self.field_suggested_index.remove(field_name)
                        self.field_suggested_index.add(substitution)
                    if self.field_suggested_primary_key is not None and field_name in self.field_suggested_primary_key:
                        index = self.field_suggested_primary_key.index(field_name)
                        self.field_suggested_primary_key[index] = substitution
        if not mode_df:
            # add columns attribute to List[dict]
            if not(isinstance(records, ListRecords)):
                records = ListRecords(records)  # this is not compatible with the inplace=True argument
            records.columns = list(self.fields_encountered.keys())
        if len(self.field_changes) > 0:
            if self.param_verbose:
                msg = "Recommended field changes: " + ", ".join({field.name: field.data_type for field in self.field_changes.values()})
                print(msg)
        if self.field_suggested_primary_key is not None:
            if not all([field_name in self.fields_encountered.keys() for field_name in self.field_suggested_primary_key]):
                self.field_suggested_primary_key = None  # cancel suggestion
        if self.field_suggested_primary_key is not None and self.field_suggested_index is not None:
            self.field_suggested_index = self.field_suggested_index - set(self.field_suggested_primary_key)
        if len(self.fields_new) > 0 and self.param_verbose:
            msg = ("The following new fields were detected: "
                   + str({field.name: field.data_type for field in self.fields_new.values()}))
            warn(msg)
            # user must call apply_new_fields_request in order to transmit new fields to CKAN
        self._extra_checks(records, fields)
        return records


class CkanDataCleanerUploadDigitalColumns(CkanDataCleanerUploadBasic):
    def __init__(self):
        super().__init__()
        self.param_replace_nan = False
        self.param_round_values = False
        self.param_rename_fields_underscore = False

    @staticmethod
    def get_class_keyword() -> str:
        return "DigitalColumns"


def default_cleaner() -> CkanDataCleanerABC:
    return CkanDataCleanerUploadBasic()


if __name__ == "__main__":
    NaN = math.nan
    date_example = datetime.datetime.today()
    timestamp_example = date_example.isoformat(ckan_timestamp_sep)

    A = {"text": "A",   "int": 1,     "number": 2,     "json": {"key": "field"},   "timestamp": timestamp_example, "test": True}
    B = {"text": "B",   "int": 1.5,   "number": 2.5,   "json": {"key": [1, 2, "A"]}, "timestamp": None, "test": None}
    C = {"text": None,  "int": None,  "number": None,  "json": {"key": [1, 2, None]}, "timestamp": pd.NaT}
    D = {"text": 1,     "int": NaN,   "number": NaN,   "json": {"key": [1, 2, NaN]}}
    E = {"text": "E",   "int": 2,     "number": 5.5,   "json": None}
    F = {"text": NaN,   "int": None,  "number": None,  "json": NaN}
    G = {"text": "G",   "int": math.inf}
    H = {"text": "H",   "extra_field": 2}

    records = [A, B, C, D, E, F, G, H]
    df = records_to_df(records)

    fields_list = [
        CkanField("text", "text"),
        CkanField("int", "int"),
        CkanField("number", "numeric"),
        CkanField("json", "json"),
        CkanField("timestamp", "timestamp"),
    ]
    fields = OrderedDict([(field_info.name, field_info) for field_info in fields_list])

    cleaner = CkanDataCleanerUploadBasic()
    auto_fields = cleaner.detect_field_types_and_subs(records, known_fields=None)
    df_cleaned = cleaner.clean_records(df, fields)
    df_warnings = cleaner.warnings
    fields_new = cleaner.fields_new
    df_records = df_cleaned.to_dict(orient="records")
    records_cleaned = cleaner.clean_records(records, fields)
    records_warnings = cleaner.warnings

    print("Done.")




#!python3
# -*- coding: utf-8 -*-
"""
The basic file format for DataStore: JSON
"""
from typing import Union, Dict
import io

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.auxiliary.list_records import ListRecords
from ckanapi_harvesters.harvesters.file_formats.file_format_abc import FileFormatABC



class JsonFileFormat(FileFormatABC):
    """
    JSON file IO using pandas.read_json.

    Reading by chunks is allowed in mode `lines=True` (reads file as one JSON object per line).
    In this case, use CLI arguments `--allow-chunks --read-kwargs lines=True`

    Recommended: use with CLI argument `--allow-chunks --read-kwargs orient=records,lines=True`

    NB: argument typ="frame" cannot be overridden for read arguments.
    """
    # NB: here, default is orient=records, lines=False for read
    #     pandas default for typ="frame" is orient=index, lines=False
    default_read_kwargs = dict(orient="records", lines=False)
    # by default, write one line per row (lines=True)
    default_write_kwargs = dict(orient="records", lines=True)

    # read -------------------
    def read_by_chunks_allowed(self) -> bool:
        return "lines" in self.read_kwargs.keys() and self.read_kwargs["lines"]

    def _get_read_kwargs(self, allow_chunks:bool=True) -> dict:
        kwargs = super()._get_read_kwargs(allow_chunks=allow_chunks)
        if self.read_by_chunks_enabled(allow_chunks=allow_chunks):
            kwargs["chunksize"] = self.chunk_size
        elif "chunksize" in kwargs.keys():
            kwargs.pop("chunksize")
        return kwargs

    def read_file(self, file_path: str, fields: Union[Dict[str, CkanField],None], allow_chunks:bool=True) -> Union[pd.DataFrame, ListRecords]:
        read_kwargs = self._get_read_kwargs(allow_chunks=allow_chunks)
        return pd.read_json(file_path, typ="frame", **read_kwargs)

    def read_buffer_full(self, buffer: io.StringIO, fields: Union[Dict[str, CkanField],None]) -> Union[pd.DataFrame, ListRecords]:
        read_kwargs = self._get_read_kwargs(allow_chunks=False)
        return pd.read_json(buffer, typ="frame", **read_kwargs)

    # write ------------------
    def write_file(self, df: pd.DataFrame, file_path: str, fields: Union[Dict[str, CkanField],None]) -> None:
        write_kwargs = self._get_write_kwargs()
        df.to_json(file_path, index=False, **write_kwargs)

    def write_in_memory(self, df: pd.DataFrame, fields: Union[Dict[str, CkanField],None]) -> bytes:
        with io.BytesIO() as buffer:
            write_kwargs = self._get_write_kwargs()
            df.to_json(buffer, index=False, **write_kwargs)
            return buffer.getvalue()

    def append_allowed(self) -> bool:
        return ("lines" in self.write_kwargs and self.write_kwargs["lines"]
                and "orient" in self.write_kwargs and self.write_kwargs["orient"] == "records")

    def append_file(self, df: Union[pd.DataFrame, ListRecords], file_path: str,
                    fields: Union[Dict[str, CkanField], None]) -> None:
        write_kwargs = self._get_write_kwargs()
        df.to_json(file_path, index=False, mode='a', **write_kwargs)

    def append_in_memory(self, stream: bytes, df: Union[pd.DataFrame, ListRecords], fields: Union[Dict[str, CkanField],None]) -> bytes:
        with io.StringIO(stream.decode("utf8")) as string_stream:
            write_kwargs = self._get_write_kwargs()
            df.to_json(string_stream, index=False, **write_kwargs)
            return string_stream.getvalue().encode("utf8")

    # misc ------------------
    def copy(self, dest=None):
        if dest is None:
            dest = JsonFileFormat(self.options_string, read_kwargs=self.read_kwargs, write_kwargs=self.write_kwargs)
        super().copy(dest=dest)
        return dest






#!python3
# -*- coding: utf-8 -*-
"""
ckanapi_harvesters
Package with helper function for CKAN requests using pandas DataFrames.
"""

time_package_loading = False

if time_package_loading:
    import time
    start_time = time.time()

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:  # Python <3.8
    from importlib_metadata import version, PackageNotFoundError

if time_package_loading:
    importlib_time = time.time()

try:
    __version__ = version("ckanapi_harvesters")
except PackageNotFoundError:
    __version__ = None

if time_package_loading:
    version_time = time.time()


import os
package_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

if time_package_loading:
    package_dir_time = time.time()

from . import auxiliary
from . import policies
from . import harvesters
from . import ckan_api
from . import builder
from . import reports

# usage shortcuts
from .auxiliary import CkanMap, RequestType, CkanField, CkanState, CkanProgressCallback, CkanCallbackLevel, CkanProgressBarType
from .policies import CkanPackageDataFormatPolicy
from .reports import CkanAdminReport
from .ckan_api import CkanApi, CkanApiParams, CKAN_API_VERSION
from .builder import BUILDER_FILE_FORMAT_VERSION
from .builder import BuilderPackage, BuilderResourceABC, BuilderDataStoreABC, BuilderDataStoreMultiABC, BuilderDataStoreFolder, RequestFileMapperIndexKeys


if time_package_loading:
    final_time = time.time()

    version_duration = version_time - start_time
    package_dir_duration = package_dir_time - version_time
    real_load_duration = final_time - package_dir_time
    overall_duration = final_time - start_time
    print(f"Package ckanapi_harvesters loaded in {overall_duration} seconds ({real_load_duration} itself, {package_dir_duration} for package_dir, {version_duration} for version)")
    # import time; start_time = time.time(); import ckanapi_harvesters; print(f"import done in {time.time() - start_time} seconds")



#!python3
# -*- coding: utf-8 -*-
"""
Call existing tests
"""
import os

from ckanapi_harvesters.ckan_api import CkanApi

from ckanapi_harvesters.builder.example.builder_example_tests_offline import run as run_tests_offline
from ckanapi_harvesters.builder.example.builder_example_patch_upload import run as run_patch_upload
from ckanapi_harvesters.builder.example.builder_example_download import run as run_download


def run(ckan:CkanApi = None):
    run_tests_offline()
    run_patch_upload(ckan=ckan)
    run_download(ckan=ckan)
    print("Tests done.")


if __name__ == '__main__':
    ckan = CkanApi(None)
    ckan.initialize_from_cli_args()
    run(ckan)




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to convert formats between database and local files.
"""
from typing import Dict, List, Union

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_model import CkanField


# user custom transformation function prototypes
def upload_function_example(df_local: Union[pd.DataFrame, List[dict]], *,
                            fields:Dict[str, CkanField]=None, file_query:str=None, total_lines_read:int=None, **kwargs) -> Union[pd.DataFrame, List[dict]]:
    return df_local

def download_function_example(df_download: pd.DataFrame, *,
                              fields:Dict[str, CkanField]=None, file_query:str=None, **kwargs) -> Union[pd.DataFrame, List[dict]]:
    return df_download


# specific examples
def simple_upload_fun(df_local: Union[pd.DataFrame, List[dict]], *,
                      fields:Dict[str, CkanField]=None, file_query:str=None, total_lines_read:int=None, **kwargs) -> Union[pd.DataFrame, List[dict]]:
    for field in df_local.columns:
        if df_local[field].dtype == pd.Timestamp:
            df_local[field] = df_local[field].apply(pd.Timestamp.isoformat)  # ISO-8601 format
    return df_local

def replace_empty_str(df_local: Union[pd.DataFrame, List[dict]], *,
                      fields:Dict[str, CkanField]=None, file_query:str=None, total_lines_read:int=None, **kwargs) -> Union[pd.DataFrame, List[dict]]:
    df_local = df_local.astype(object)  # set all columns to object dtype
    df_local.replace("", None, inplace=True)
    return df_local




#!python3
# -*- coding: utf-8 -*-
"""
Methods to load authentication credentials (user, password)
"""
from typing import Union, Tuple, Dict
import getpass
from warnings import warn
import os
import argparse
import io

from ckanapi_harvesters.auxiliary.ckan_defs import environ_keyword
from ckanapi_harvesters.auxiliary.ckan_errors import LoginFileError
from ckanapi_harvesters.auxiliary.path import sanitize_path, path_rel_to_dir


class Login:
    LOGIN_FILE_ENVIRON: Union[str, None] = None

    def __init__(self, username:str=None, password:str=None, login_file:str=None,
                 auto_load:bool=True, login_file_environ:str=None):
        self._username:Union[str,None] = username
        self._password:Union[str,None] = password
        self.login_file: str = login_file  # path to a file containing a valid API key in the first line of text (optional)
        self.LOGIN_FILE_ENVIRON = login_file_environ
        if auto_load:
            if self.is_empty() and login_file is not None:
                self.load_from_file(error_not_found=False)
            if self.is_empty():
                if self.load_from_environ(error_not_found=False, empty_warning=False):
                    print(f"API key loaded from file: {self.apikey_file}")

    def __del__(self):
        self.clear()

    def __copy__(self):
        return self.copy()

    def copy(self, *, dest=None):
        if dest is None:
            dest = Login()
        dest.login_file = self.login_file
        dest._username = self._username
        dest._password = self._password
        return dest

    def clear(self) -> None:
        self._username = None
        self._password = None

    def is_empty(self):
        return self._username is None or self._password is None

    def __str__(self):
        if self.is_empty():
            return "None"
        elif self._username == "" and self._password == "":
            return "<empty string>"
        elif self._username == "":
            return "<empty user>"
        elif self._password == "":
            return "<empty password>"
        else:
            return "*****"

    @property
    def username(self) -> Union[str,None]:
        return self._username
    @username.setter
    def username(self, value:Union[str,None]):
        self._username = value
    @property
    def password(self) -> Union[str,None]:
        return self._password
    @password.setter
    def password(self, value:Union[str,None]):
        self._password = value

    def get_default_login_file(self) -> Union[str, None]:
        if self.LOGIN_FILE_ENVIRON is None:
            return None
        env_apikey_file = os.environ.get(self.LOGIN_FILE_ENVIRON)   # "CKAN_API_KEY_FILE"
        if env_apikey_file:
            assert not env_apikey_file.strip().lower() == environ_keyword  # this value would create an infinite loop
            return sanitize_path(env_apikey_file)
        else:
            return None

    def load_from_environ(self, *, error_not_found:bool=False, empty_warning:bool=True) -> bool:
        """
        Load login from environment variables, by order of priority:

        - LOGIN_FILE_ENVIRON: path to a file containing the username, password

        :param error_not_found: raise an error if the API key file was not found
        :return:
        """
        default_login_file = self.get_default_login_file()
        if default_login_file is not None:
            return self.load_from_file(default_login_file, error_not_found=error_not_found)
        else:
            if empty_warning:
                msg = f"No API key was found in the environment variable {self.LOGIN_FILE_ENVIRON}"
                warn(msg)
            return False

    def load_from_file(self, login_file:str=None, *, base_dir:str=None, error_not_found:bool=True) -> bool:
        """
        Load the credentials from file.
        The file should contain username in first line and password in second line.

        :param login_file: path to the API key file. The following keywords are accepted:
            - "environ": the API key will be looked up in the environment variable with load_from_environ
        :param base_dir: base directory to find the API key file, if a relative path is provided
        :param error_not_found: option to raise an exception if the API key file is not found
        :return:
        """
        if login_file is None:
            login_file = self.login_file
        login_file = path_rel_to_dir(login_file, base_dir=base_dir, keyword_exceptions={environ_keyword})
        if login_file is None:
            raise LoginFileError('login_file is required')
        login_keyword = login_file.strip().lower()
        if login_keyword == environ_keyword:
            return self.load_from_environ(error_not_found=error_not_found)
        if not(os.path.isfile(login_file)) and not error_not_found:
            msg = f"Login file does not exist: {login_file}"
            warn(msg)
            return False
        with open(login_file, 'r') as f:
            self.username = f.readline().strip()
            self.password = f.readline().rstrip('\n')
            f.close()
        self.login_file = login_file
        return True

    def input(self):
        """
        Prompt the user to input the login credentials in the console window.

        :return:
        """
        value = input("Please enter user name: ")
        self.username = value
        value = getpass.getpass("Please enter password: ")
        self.password = value

    @staticmethod
    def _setup_cli_parser(parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="Login credentials initialization", add_help=False)
        parser.add_argument("--login-file", type=str,
                            help="Path to a text file containing login credentials for authentification (user, password)")
        parser.add_argument("--login-user", type=str,
                            help="Login user name (prefer using a file)")
        parser.add_argument("--login-password", type=str,
                            help="Login user password (prefer using a file)")
        return parser

    def print_help_cli(self, display:bool=True) -> str:
        parser = self._setup_cli_parser()
        if display:
            parser.print_help()
        with io.StringIO() as stream:
            parser.print_help(stream)
            return stream.getvalue()

    def _cli_args_apply(self, args: argparse.Namespace, *, base_dir: str = None, error_not_found: bool = True) -> None:
        if args.login_file is not None:
            self.load_from_file(args.login_file, base_dir=base_dir, error_not_found=error_not_found)
        if args.login_user is not None:
            self.username = args.login_user
        if args.login_password is not None:
            self.password = args.login_password

    def to_tuple(self) -> Tuple[str,str]:
        return self.username, self.password

    @staticmethod
    def from_tuple(values: Tuple[str,str]) -> "Login":
        login = Login(*values)
        return login

    def to_dict(self) -> Dict[str,str]:
        return {"username": self.username, "password": self.password}

    @staticmethod
    def from_dict(values: Dict[str,str]) -> "Login":
        login = Login(**values)
        return login





#!python3
# -*- coding: utf-8 -*-
"""
Tests to perform after the example package was uploaded
"""
import os
import json

from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.ckan_api import CkanApi

from ckanapi_harvesters.builder.example import example_package_xls

def run():
    BuilderPackage.unlock_external_code_execution()  # comment to test if the safety feature is enabled

    mdl = BuilderPackage.from_excel(example_package_xls)

    ckan = CkanApi()
    ckan = mdl.init_ckan(ckan)
    # mdl.update_ckan_map(ckan)
    # map_from_mdl = ckan.map.copy()

    # Test re-encoding the Excel file from the loaded model
    example_package_xls_reencoded = os.path.abspath("builder_package_example-reencoded.xlsx")
    mdl.to_excel(example_package_xls_reencoded)

    # export mdl to dict and re-import
    base_dir = os.path.abspath(".")
    mdl_dict = mdl.to_dict(base_dir=base_dir)
    mdl_from_dict = BuilderPackage.from_dict(mdl_dict, base_dir=base_dir)
    mdl_dict_bis = mdl_from_dict.to_dict(base_dir=base_dir)
    assert(mdl_dict == mdl_dict_bis)

    # test json serialization
    example_package_json_reencoded = os.path.abspath("builder_package_example-reencoded.json")
    with open(example_package_json_reencoded, "w") as json_file:
        json.dump(mdl_dict, json_file, indent=4)

    # test copy constructors
    ckan_copy = ckan.copy()
    mdl_copy = mdl.copy()
    mdl_copy_dict = mdl_copy.to_dict(base_dir=base_dir)
    assert(mdl_dict == mdl_copy_dict)

    print("Tests done.")

if __name__ == '__main__':
    run()
